from compas.geometry import Box
from compas.geometry import Line
from compas.geometry import Plane
from compas.geometry import distance_point_point
from compas.geometry import intersection_line_plane
from compas_timber.connections import InterfaceLocation

from .wall_populator import BeamDefinition


def _get_furthest_line(lines, point):
    furthest_line = lines[0]  # just to not start with None but one line will always be the furthest
    max_distance = -float("inf")

    for line in lines:
        start_distance = distance_point_point(line.start, point)
        end_distance = distance_point_point(line.end, point)

        max_line_projection = max(start_distance, end_distance)
        if max_line_projection > max_distance:
            max_distance = max_line_projection
            furthest_line = line
    return furthest_line


class LDetailBase(object):
    def adjust_segments_main(self, interface, wall, config_set, perimeter_segments):
        # top and bottom segments are shortened/extended to the intersection plane
        # front or back (depending on the end at which the deailt is) segment are moved to the end of the interface
        # shorten top and bottom segments to the interface
        interface_plane = interface.as_plane()
        top_segment = perimeter_segments["top"]
        bottom_segment = perimeter_segments["bottom"]
        intersection_top = intersection_line_plane(top_segment, interface_plane)
        intersection_bottom = intersection_line_plane(bottom_segment, interface_plane)

        if interface.interface_type == InterfaceLocation.BACK:
            perimeter_segments["top"] = Line(top_segment.start, intersection_top)
            perimeter_segments["bottom"] = Line(intersection_bottom, bottom_segment.end)

        elif interface.interface_type == InterfaceLocation.FRONT:
            perimeter_segments["top"] = Line(intersection_top, top_segment.end)
            perimeter_segments["bottom"] = Line(bottom_segment.start, intersection_bottom)

    def adjust_segments_cross(self, interface, wall, config_set, perimeter_segments):
        distance_a = distance_point_point(interface.interface_polyline[1], wall.baseline.midpoint)
        distance_b = distance_point_point(interface.interface_polyline[2], wall.baseline.midpoint)
        if distance_a > distance_b:
            outer_point = interface.interface_polyline[1]
        else:
            outer_point = interface.interface_polyline[2]
        edge_plane = Plane(outer_point, wall.baseline.direction)  # TODO: using interface.frame.zaxis instead
        top_segment = perimeter_segments["top"]
        bottom_segment = perimeter_segments["bottom"]
        bottom_point = intersection_line_plane(bottom_segment, edge_plane)
        top_point = intersection_line_plane(top_segment, edge_plane)
        if interface.interface_type == InterfaceLocation.FRONT:
            bottom_segment = Line(bottom_segment.start, bottom_point)
            top_segment = Line(top_point, top_segment.end)
        elif interface.interface_type == InterfaceLocation.BACK:
            bottom_segment = Line(bottom_point, bottom_segment.end)
            top_segment = Line(top_segment.start, top_point)

        perimeter_segments["top"] = top_segment
        perimeter_segments["bottom"] = bottom_segment


class TDetailBase(object):
    def adjust_segments_main(self, interface, wall, config_set, perimeter_segments):
        # top and bottom segments are shortened/extended to the intersection plane
        interface_plane = interface.as_plane()
        top_segment = perimeter_segments["top"]
        bottom_segment = perimeter_segments["bottom"]
        intersection_top = intersection_line_plane(top_segment, interface_plane)
        intersection_bottom = intersection_line_plane(bottom_segment, interface_plane)

        if interface.interface_type == InterfaceLocation.BACK:
            print("shortening top and bottom back")
            perimeter_segments["top"] = Line(top_segment.start, intersection_top)
            perimeter_segments["bottom"] = Line(intersection_bottom, bottom_segment.end)

        elif interface.interface_type == InterfaceLocation.FRONT:
            print("shortening top and bottom front")
            perimeter_segments["top"] = Line(intersection_top, top_segment.end)
            perimeter_segments["bottom"] = Line(bottom_segment.start, intersection_bottom)

    def adjust_segments_cross(self, *args, **kwargs):
        # top and bottom are not modified
        # internal segments are created on either side of the interface
        pass


class LConnectionDetailB(LDetailBase):
    """
    Parameters
    ----------
    interface : :class:`compas_timber.connections.WallToWallInterface`
    """

    def get_detail_obb_main(self, interface, config_set, wall):
        xsize = config_set.beam_width * 2.0  # xsize
        ysize = interface.interface_polyline.lines[0].length
        zsize = wall.thickness
        box_frame = interface.frame.copy()
        box_frame.point += interface.frame.xaxis * xsize * 0.5
        box_frame.point += interface.frame.yaxis * ysize * 0.5
        box_frame.point += interface.frame.zaxis * zsize * 0.5
        return Box(xsize, ysize, zsize, frame=box_frame)

    def get_detail_obb_cross(self, interface, config_set, wall):
        if interface.interface_type == InterfaceLocation.FRONT:
            parallel_to_interface = wall.baseline.direction * -1.0
        else:
            parallel_to_interface = wall.baseline.direction
        xsize = wall.thickness
        ysize = interface.interface_polyline.lines[0].length
        zsize = 2 * config_set.beam_width + wall.thickness  # a bit bigger than needs to be
        box_frame = interface.frame.copy()
        box_frame.point += interface.frame.xaxis * xsize * 0.5
        box_frame.point += interface.frame.yaxis * ysize * 0.5
        box_frame.point += parallel_to_interface * zsize * 0.5
        return Box(xsize, ysize, zsize, frame=box_frame)

    def create_elements_cross(self, interface, wall, config_set):
        if interface.interface_type == InterfaceLocation.FRONT:
            left_vertical = interface.interface_polyline.lines[0]
            parallel_to_interface = wall.baseline.direction * -1.0
        else:
            left_vertical = interface.interface_polyline.lines[2]
            parallel_to_interface = wall.baseline.direction

        perpendicular_to_interface = interface.frame.xaxis

        edge_offset = config_set.beam_width * 0.5 + config_set.edge_stud_offset
        edge_beam_line = left_vertical.translated(parallel_to_interface * edge_offset)
        edge_beam = BeamDefinition(edge_beam_line, config_set.beam_width, wall.thickness, normal=perpendicular_to_interface, type="detail_edge")

        between_edge = edge_beam_line.translated(perpendicular_to_interface * 0.5 * config_set.beam_width)
        between_edge.translate(parallel_to_interface * 0.5 * config_set.beam_width)
        between_beam = BeamDefinition(between_edge, config_set.beam_width, wall.thickness, normal=parallel_to_interface, type="detail")
        return [between_beam, edge_beam]

    def create_elements_main(self, interface, wall, config_set):
        # create a beam (definition) as wide and as high as the wall
        # it should be flush agains the interface
        polyline = interface.interface_polyline
        beam_zaxis = interface.frame.normal
        reference_edge = polyline.lines[0].translated(interface.frame.xaxis * config_set.beam_width * 0.5)
        edge_beam = BeamDefinition(reference_edge, config_set.beam_width, wall.thickness, normal=beam_zaxis, type="detail_edge")
        return [edge_beam]


class LConnectionDetailA(LDetailBase):
    """
    Parameters
    ----------
    interface : :class:`compas_timber.connections.WallToWallInterface`
    """

    def get_detail_obb_main(self, interface, config_set, wall):
        xsize = config_set.beam_width * 2.0  # xsize
        ysize = interface.interface_polyline.lines[0].length
        zsize = wall.thickness
        box_frame = interface.frame.copy()
        box_frame.point += interface.frame.xaxis * xsize * 0.5
        box_frame.point += interface.frame.yaxis * ysize * 0.5
        box_frame.point += interface.frame.zaxis * zsize * 0.5
        return Box(xsize, ysize, zsize, frame=box_frame)

    def get_detail_obb_cross(self, interface, config_set, wall):
        if interface.interface_type == InterfaceLocation.FRONT:
            parallel_to_interface = wall.baseline.direction * -1.0
        else:
            parallel_to_interface = wall.baseline.direction
        xsize = wall.thickness
        ysize = interface.interface_polyline.lines[0].length
        zsize = 3 * config_set.beam_width + wall.thickness  # a bit bigger than needs to be
        box_frame = interface.frame.copy()
        box_frame.point += interface.frame.xaxis * xsize * 0.5
        box_frame.point += interface.frame.yaxis * ysize * 0.5
        box_frame.point += parallel_to_interface * zsize * 0.5
        return Box(xsize, ysize, zsize, frame=box_frame)

    def create_elements_cross(self, interface, wall, config_set):
        if interface.interface_type == InterfaceLocation.FRONT:
            parallel_to_interface = wall.baseline.direction * -1.0  # this should always point from the wall outwards direction of the interface
        else:
            parallel_to_interface = wall.baseline.direction

        vertical_lines = [interface.interface_polyline.lines[0], interface.interface_polyline.lines[2]]
        edge_vertical = _get_furthest_line(vertical_lines, wall.baseline.midpoint)

        perpendicular_to_interface = interface.frame.xaxis

        edge_offset = config_set.beam_width * 0.5 + config_set.edge_stud_offset
        edge_beam_line = edge_vertical.translated(parallel_to_interface * edge_offset)
        edge_beam = BeamDefinition(edge_beam_line, config_set.beam_width, wall.thickness, normal=perpendicular_to_interface, type="detail_edge")

        other_edge_line = edge_beam_line.translated(parallel_to_interface * 1.0 * (wall.thickness + config_set.beam_width))
        other_beam = BeamDefinition(other_edge_line, config_set.beam_width, wall.thickness, normal=perpendicular_to_interface, type="detail")

        between_edge = edge_beam_line.translated(perpendicular_to_interface * 0.5 * config_set.beam_width)
        between_edge.translate(parallel_to_interface * 0.5 * config_set.beam_width)
        between_beam = BeamDefinition(between_edge, config_set.beam_width, wall.thickness, normal=parallel_to_interface, type="detail")
        return [between_beam, edge_beam, other_beam]

    def create_elements_main(self, interface, wall, config_set):
        # create a beam (definition) as wide and as high as the wall
        # it should be flush agains the interface
        polyline = interface.interface_polyline
        beam_zaxis = interface.frame.normal
        reference_edge = polyline.lines[0].translated(interface.frame.xaxis * config_set.beam_width * 0.5)
        edge_beam = BeamDefinition(reference_edge, config_set.beam_width, wall.thickness, normal=beam_zaxis, type="detail_edge")
        return [edge_beam]


class TConnectionDetailA(TDetailBase):
    """
    Parameters
    ----------
    interface : :class:`compas_timber.connections.WallToWallInterface`
    """

    def get_detail_obb_main(self, interface, config_set, wall):
        xsize = config_set.beam_width * 2.0  # xsize
        ysize = interface.interface_polyline.lines[0].length
        zsize = wall.thickness
        box_frame = interface.frame.copy()
        box_frame.point += interface.frame.xaxis * xsize * 0.5
        box_frame.point += interface.frame.yaxis * ysize * 0.5
        box_frame.point += interface.frame.zaxis * zsize * 0.5
        return Box(xsize, ysize, zsize, frame=box_frame)

    def get_detail_obb_cross(self, interface, config_set, wall):
        xsize = wall.thickness  # xsize
        ysize = interface.interface_polyline.lines[0].length
        zsize = wall.thickness
        box_frame = interface.frame.copy()
        box_frame.point += interface.frame.xaxis * xsize * 0.5
        box_frame.point += interface.frame.yaxis * ysize * 0.5
        box_frame.point += interface.frame.zaxis * zsize * 0.5
        return Box(xsize, ysize, zsize * 1.5, frame=box_frame)

    def create_elements_cross(self, interface, wall, config_set):
        # create a beam (definition) as wide and as high as the wall
        # it should be flush agains the interface
        polyline = interface.interface_polyline
        top_midpoint = polyline.lines[1].midpoint
        bottom_midpoint = polyline.lines[3].midpoint
        axis = Line(top_midpoint, bottom_midpoint)
        perpendicular_to_interface = interface.frame.xaxis
        edge_beam = BeamDefinition(axis, config_set.beam_width, wall.thickness, normal=perpendicular_to_interface, type="detail")

        return [edge_beam]

    def create_elements_main(self, interface, wall, config_set):
        # create a beam (definition) as wide and as high as the wall
        # it should be flush agains the interface
        polyline = interface.interface_polyline
        beam_zaxis = interface.frame.normal
        reference_edge = polyline.lines[0].translated(interface.frame.xaxis * config_set.beam_width * 0.5)
        edge_beam = BeamDefinition(reference_edge, config_set.beam_width, wall.thickness, normal=beam_zaxis, type="detail_edge")
        return [edge_beam]



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper
import Rhino
import System
from compas.geometry import Brep

from compas_timber.elements import FastenerTimberInterface


class FastenerTimberInterfaceComponent(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(
        self,
        outline: Rhino.Geometry.Polyline,
        thickness: float,
        drill_points: System.Collections.Generic.List[Rhino.Geometry.Point3d],
        drill_diameters: System.Collections.Generic.List[float],
        extra_shapes: System.Collections.Generic.List[Rhino.Geometry.Brep],
        features: System.Collections.Generic.List[object],
    ):
        holes = []
        if drill_points and drill_diameters:
            if len(drill_points) == len(drill_diameters):
                for pt, dia in zip(drill_points, drill_diameters):
                    holes.append({"point": pt, "diameter": dia, "vector": None, "through": True})
            elif len(drill_diameters) == 1:
                for pt in drill_points:
                    holes.append({"point": pt, "diameter": drill_diameters[0], "vector": None, "through": True})
            else:
                raise Warning("Number of diameters must either match the number of points or be a single value")
        feat_list = []
        if features:
            for feature in features:
                if feature:
                    if feature.elements:
                        ghenv.Component.AddRuntimeMessage(
                            Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning,
                            "Features in the Fastener Timber Interface are applied by joints. Elements in feature definitions will be ignored",
                        )
                    feat_list.append(feature)
        outline_points = [pt for pt in outline] if outline else None
        interface = FastenerTimberInterface(
            outline_points,
            thickness,
            holes,
            shapes=[Brep.from_native(brep) for brep in extra_shapes] if extra_shapes else [],
            features=feat_list,
        )
        return interface



from __future__ import print_function

import os


__author__ = ["Chen Kasirer"]
__copyright__ = "Gramazio Kohler Research"
__license__ = "MIT License"
__email__ = "kasirer@arch.ethz.ch"
__version__ = "0.1.0"


HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))


__all__ = ["HOME", "DATA", "DOCS"]



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper
import System
import rhinoscriptsyntax as rs


from compas_rhino.conversions import frame_to_rhino
from timber_design.ghpython.ghcomponent_helpers import list_input_valid_cpython


class ShowElementFaces(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, element: System.Collections.Generic.List[object], ref_side_index: System.Collections.Generic.List[int]):
        if not list_input_valid_cpython(ghenv, element, "Element"):
            return
        self.pl = []
        self.txt = []
        self.ht = []
        srfs = []
        if not ref_side_index:
            ref_side_index = [0]
        if not len(ref_side_index) == len(element):
            ref_side_index = [ref_side_index[0] for _ in element]

        for b, i in zip(element, ref_side_index):
            srfs.append(self.get_srf(b, i))
            ht = 1000
            for side_index in range(len(b.ref_sides)):
                surface = b.side_as_surface(side_index)
                ht = min([ht, surface.xsize / 6.0, surface.ysize / 6.0])
            for side_index in range(len(b.ref_sides)):
                surface = b.side_as_surface(side_index)
                frame = b.ref_sides[side_index]
                frame.point = surface.point_at(0, ht / 4.0)

                self.pl.append(frame_to_rhino(frame))
                self.txt.append(surface.name)
                self.ht.append(ht)

        return srfs

    def get_srf(self, element, ref_side_index):
        face = element.side_as_surface(ref_side_index)
        rh_srf = rs.AddPlaneSurface(frame_to_rhino(face.frame), face.xsize, face.ysize)
        return rh_srf

    def DrawViewportWires(self, arg):
        if ghenv.Component.Locked:
            return
        col = System.Drawing.Color.FromArgb(255, 255, 255, 255)
        # https://developer.rhino3d.com/api/RhinoCommon/html/M_Rhino_Display_DisplayPipeline_Draw3dText_5.htm
        for p, t, h in zip(self.pl, self.txt, self.ht):
            arg.Display.Draw3dText(t, col, p, h, "Verdana", True, False)



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper
from compas.scene import Scene
from System.Windows.Forms import ToolStripSeparator

from timber_design.ghpython.ghcomponent_helpers import rename_gh_output


class ShowElementsByCategory(Grasshopper.Kernel.GH_ScriptInstance):
    def __init__(self):
        super(ShowElementsByCategory, self).__init__()
        ghenv.Component.Params.Output[0].NickName == "elements"
        self.element_type = None

    def RunScript(self, model):
        if model is None:
            self.names = set()
            ghenv.Component.Message = "Please connect Model"
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Please connect Model")
            return None

        self.names = set(element.attributes.get("category") for element in model.elements())
        self.names.discard(None)

        if not self.element_type:
            ghenv.Component.Message = "Select element type from context menu (right click)"
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Select element type from context menu (right click)")
            return None

        ghenv.Component.Message = f"Category: {self.element_type}"

        scene = Scene()
        for element in model.elements():
            if element.attributes.get("category") == self.element_type:
                scene.add(element.geometry)
        return scene.draw()

    def AppendAdditionalMenuItems(self, menu):
        menu.Items.Add(ToolStripSeparator())
        for name in self.names:
            item = menu.Items.Add(name, None, self.on_item_click)
            if self.element_type and name == self.element_type:
                item.Checked = True

    def on_item_click(self, sender, event_info):
        self.element_type = sender.Text
        ghenv.Component.ExpireSolution(True)



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper
import System

from compas_timber.connections import JointTopology
from timber_design.workflow import LConnectionDetailA
from timber_design.workflow import TConnectionDetailA
from timber_design.workflow import WallPopulatorConfigurationSet
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class WallPopulatorConfigSetComponent(Grasshopper.Kernel.GH_ScriptInstance):
    DEFAULT_DETAILS = {JointTopology.TOPO_L: LConnectionDetailA(), JointTopology.TOPO_T: TConnectionDetailA()}

    def RunScript(
        self,
        stud_spacing: float,
        beam_width: float,
        sheeting_outside: float,
        sheeting_inside: float,
        lintel_posts: bool,
        edge_stud_offset: float,
        custom_dimensions: System.Collections.Generic.List[object],
        joint_overrides: System.Collections.Generic.List[object],
    ):
        """default values for stud spacing and beam width"""
        if not item_input_valid_cpython(ghenv, stud_spacing, "Stud Spacing"):
            unit_system = Rhino.RhinoDoc.ActiveDoc.ModelUnitSystem
            if unit_system == Rhino.UnitSystem.Meters:
                stud_spacing = 0.625
            elif unit_system == Rhino.UnitSystem.Centimeters:
                stud_spacing = 62.5
            elif unit_system == Rhino.UnitSystem.Millimeters:
                stud_spacing = 625.0
            else:
                ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Error, "No default stud_spacing for freedom units, silly billy!")

        if not item_input_valid_cpython(ghenv, beam_width, "Beam Width"):
            unit_system = Rhino.RhinoDoc.ActiveDoc.ModelUnitSystem
            if unit_system == Rhino.UnitSystem.Meters:
                beam_width = 0.06
            elif unit_system == Rhino.UnitSystem.Centimeters:
                beam_width = 6.0
            elif unit_system == Rhino.UnitSystem.Millimeters:
                beam_width = 60.0
            else:
                ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Error, "No default beam_width for freedom units, silly billy!")

        dims = {}
        for item in custom_dimensions:
            for key, val in item.items():
                dims[key] = val

        config_set = WallPopulatorConfigurationSet.default(stud_spacing, beam_width)
        config_set.connection_details = self.DEFAULT_DETAILS

        if dims is not None:
            config_set.custom_dimensions = dims
        if sheeting_inside is not None:
            config_set.sheeting_inside = sheeting_inside
        if sheeting_outside is not None:
            config_set.sheeting_outside = sheeting_outside
        if lintel_posts is not None:
            config_set.lintel_posts = lintel_posts
        if edge_stud_offset is not None:
            config_set.edge_stud_offset = edge_stud_offset
        if joint_overrides is not None:
            config_set.joint_overrides = joint_overrides
        return config_set



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper
import System
from compas_rhino.conversions import point_to_rhino

from compas_timber.utils import intersection_line_line_param
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class ShowJointTypes(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, model):
        self.pt = []
        self.txt = []

        if not item_input_valid_cpython(ghenv, model, "model"):
            return

        for joint in model.joints:
            line_a, line_b = joint.elements[0].centerline, joint.elements[1].centerline
            [p1, _], [p2, _] = intersection_line_line_param(line_a, line_b, float("inf"), False, 1e-3)
            p1 = point_to_rhino(p1)
            p2 = point_to_rhino(p2)

            self.pt.append((p2 + p1) / 2)
            self.txt.append(joint.__class__.__name__)

    def DrawViewportWires(self, arg):
        if ghenv.Component.Locked:
            return
        col = System.Drawing.Color.FromArgb(255, 0, 0, 0)
        # https://developer.rhino3d.com/api/RhinoCommon/html/M_Rhino_Display_DisplayPipeline_Draw2dText_5.htm
        for p, t in zip(self.pt, self.txt):
            arg.Display.Draw2dText(t, col, p, True, 12, "Verdana")



# r: compas_timber>=1.0.0
# flake8: noqa
import Grasshopper
import System
from ghpythonlib.treehelpers import tree_to_list

from compas_timber.planning import BeamStock as CTBeamStock
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class BeamStock(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, length: System.Collections.Generic.List[float], cross_section: Grasshopper.DataTree[float]):
        if not item_input_valid_cpython(ghenv, length, "Length"):
            return
        if not item_input_valid_cpython(ghenv, cross_section, "CrossSection"):
            return

        # Convert inputs from tree to list
        length = list(length)
        cross_section = tree_to_list(cross_section)

        # If user passes a flat list, nest it
        if all(isinstance(x, float) for x in cross_section):
            cross_section = [cross_section]

        for cs in cross_section:
            if len(cs) != 2:
                ghenv.Component.AddRuntimeMessage(
                    Grasshopper.Kernel.GH_RuntimeMessageLevel.Error,
                    "CrossSection must be a tuple of two numeric values representing height and width.",
                )
                return

        # Duplicate data if necessary
        N = len(cross_section)
        if len(length) != N:
            for _ in range(N - len(length)):
                length.append(length[0])

        # Create Stock objects
        stocks = []
        for l, cs in zip(length, cross_section):
            stock = CTBeamStock(length=l, cross_section=cs)
            stocks.append(stock)

        return stocks



# r: timber_design>=0.1.0
# venv: td_migration
"""Creates an Model"""

import Grasshopper
import Rhino
import System
from compas.scene import Scene
from compas.tolerance import TOL
from compas.tolerance import Tolerance

# from timber_design.workflow import WallPopulator - breaks the GH Component
from compas_timber.elements import Beam
from compas_timber.elements import Plate
from compas_timber.errors import FeatureApplicationError
from compas_timber.model import TimberModel

from timber_design.ghpython import error
from timber_design.ghpython import warning
from timber_design.workflow import DebugInfomation
from timber_design.workflow import JointRuleSolver

# workaround for https://github.com/gramaziokohler/compas_timber/issues/280
TOL.absolute = 1e-6


class ModelComponent(Grasshopper.Kernel.GH_ScriptInstance):
    @property
    def component(self):
        return ghenv.Component  # type: ignore

    def RunScript(
        self,
        Elements: System.Collections.Generic.List[object],
        Containers: System.Collections.Generic.List[object],
        JointRules: System.Collections.Generic.List[object],
        Features: System.Collections.Generic.List[object],
        MaxDistance: float,
        CreateGeometry: bool,
    ):
        # this used to be default behavior in Rhino7.. I think..
        Elements = Elements or []
        Containers = Containers or []
        JointRules = JointRules or []
        Features = Features or []

        if not Elements:
            warning(self.component, "Input parameter Beams failed to collect data")
        if not JointRules:
            warning(self.component, "Input parameter JointRules failed to collect data")
        if not (Elements or Containers):  # shows beams even if no joints are found
            return
        if MaxDistance is None:
            MaxDistance = TOL.ABSOLUTE  # compared to calculted distance, so shouldn't be just 0.0

        tol = self.get_tol()
        Model = TimberModel(tolerance=tol)
        debug_info = DebugInfomation()

        ##### Adding elements #####
        self.add_elements_to_model(Model, Elements, Containers)

        ##### Wall populating #####
        handled_pairs, wall_joints = self.handle_populators(Model, Containers, MaxDistance)

        ##### Handle joinery #####
        # checks elements compatibility and generates Joints
        JointRules = [j for j in JointRules if j is not None]

        if JointRules:
            solver = JointRuleSolver(JointRules, max_distance=MaxDistance)
            joint_errors, _ = solver.apply_rules_to_model(Model)  # TODO: figure out best way to pass out unjoined_clusters

            for je in joint_errors:
                debug_info.add_joint_error(je)

        # applies extensions and features resulting from joints
        bje = Model.process_joinery()
        if bje:
            debug_info.add_joint_error(bje)

        ##### Handle user features #####
        if Features:
            feature_errors = self.handle_features(Features)
            debug_info.add_feature_error(feature_errors)

        ##### Visualization #####
        Geometry, errors = self.handle_geometry(Model, CreateGeometry)
        for geo_error in errors:
            debug_info.add_feature_error(geo_error)

        ##### Error Handling #####
        if debug_info.has_errors:
            warning(self.component, "Error found during joint creation. See DebugInfo output for details.")

        return Model, Geometry, debug_info

    def get_tol(self):
        units = Rhino.RhinoDoc.ActiveDoc.GetUnitSystemName(True, True, True, True)
        if units == "m":
            return Tolerance(unit="M", absolute=1e-6, relative=1e-6)
        elif units == "mm":
            return Tolerance(unit="MM", absolute=1e-3, relative=1e-3)
        else:
            error(self.component, f"Unsupported unit: {units}")
            return

    def add_elements_to_model(self, model, elements, containers):
        """Adds elements to the model and groups them by slab."""
        elements = [e for e in elements if e is not None]
        for element in elements:
            element.reset()
            model.add_element(element)

        containers = [c for c in containers if c is not None]

        for index, c_def in enumerate(containers):
            slab = c_def.slab
            model.add_group_element(slab, name=slab.name + str(index))

    def handle_populators(self, model, containers, max_distance):
        # Handle wall populators
        model.connect_adjacent_walls()
        config_sets = [c_def.config_set for c_def in containers]
        populators = []
        if any(config_sets):
            populators = WallPopulator.from_model(model, config_sets)

        handled_pairs = []
        wall_joints = []
        for populator, slab in zip(populators, list(model.slabs)):
            elements = populator.create_elements()
            model.add_elements(elements, parent=slab.name)
            joint_definitions = populator.create_joints(elements, max_distance)
            wall_joints.extend(joint_definitions)
            for j_def in joint_definitions:
                element_a, element_b = j_def.elements
                handled_pairs.append({element_a, element_b})

        return handled_pairs, wall_joints

    def handle_features(self, features):
        feature_errors = []
        features = [f for f in features if f is not None]
        for f_def in features:
            if not f_def.elements:
                warning(self.component, "Features defined in model must have elements defined. Features without elements will be ignored")
                continue

            for element in f_def.elements:
                try:
                    element.add_features(f_def.feature_from_element(element))
                except FeatureApplicationError as ex:
                    feature_errors.append(ex)
        return feature_errors

    def handle_geometry(self, Model, CreateGeometry):
        scene = Scene()
        errors = []
        for element in Model.elements():
            if CreateGeometry:
                scene.add(element.geometry)
                if getattr(element, "debug_info", False):
                    errors.append(element.debug_info)
            else:
                if isinstance(element, Beam):
                    scene.add(element.blank)
                elif isinstance(element, Plate):
                    scene.add(element.shape)
                else:
                    scene.add(element.geometry)
        return scene.draw(), errors



from compas.tolerance import TOL
from compas_timber.analyzers import Cluster
from compas_timber.analyzers import MaxNCompositeAnalyzer
from compas_timber.connections import JointTopology
from compas_timber.connections import LMiterJoint
from compas_timber.connections import PlateMiterJoint
from compas_timber.connections import PlateTButtJoint
from compas_timber.connections import TButtJoint
from compas_timber.connections import XLapJoint
from compas_timber.errors import BeamJoiningError
from compas_timber.utils import intersection_line_line_param


class CollectionDef(object):
    """TODO: this should be removed since it's essentially a list."""

    def __init__(self, objs):
        objs = [_ for _ in objs if _]

        self.objs = objs
        self.keys_map = {}

        for i, obj in enumerate(objs):
            self.keys_map[i] = obj

    def __str__(self):
        return "Collection with %s items." % len(self.objs)


class ContainerDefinition(object):
    """Holds a pair of panel and its configuration set if available."""

    def __init__(self, panel, config_set=None):
        self.panel = panel
        self.config_set = config_set


class JointRuleSolver(object):
    """A class that holds a set of rules for joining elements.

    It can be used to apply the rules to a model and create joints based on the rules.

    Parameters
    ----------
    rules : list(:class:`~compas_timber.connections.JointRule`)
        A list of rules to apply to the model.
    use_default_topo : bool, optional
        Whether to use the default topology rules. Defaults to False.
    max_distance : float, optional
        The maximum distance to consider two elements as intersecting. Defaults to TOL.absolute.
        This is the general model.max_distance and will be overriden by individual rule max_distances.

    """

    def __init__(self, rules, use_default_topo=False, max_distance=TOL.absolute):
        self.rules = JointRuleSolver._sort_rules(rules, use_default_topo)
        self.use_default_topo = use_default_topo
        self.max_distance = max_distance
        self.clusters = []
        self.joining_errors = []

    @staticmethod
    def _sort_rules(rules, use_default_topo=False):
        """Sorts the rules by their priority."""
        direct_rules = []
        category_rules = []
        if use_default_topo:
            topo_rules = {
                JointTopology.TOPO_L: TopologyRule(JointTopology.TOPO_L, LMiterJoint),
                JointTopology.TOPO_T: TopologyRule(JointTopology.TOPO_T, TButtJoint),
                JointTopology.TOPO_X: TopologyRule(JointTopology.TOPO_X, XLapJoint),
                JointTopology.TOPO_EDGE_EDGE: TopologyRule(JointTopology.TOPO_EDGE_EDGE, PlateMiterJoint),
                JointTopology.TOPO_EDGE_FACE: TopologyRule(JointTopology.TOPO_EDGE_FACE, PlateTButtJoint),
            }
        else:
            topo_rules = {}
        for rule in rules:
            if isinstance(rule, DirectRule):
                direct_rules.append(rule)
            if isinstance(rule, CategoryRule):
                category_rules.append(rule)
            if isinstance(rule, TopologyRule):
                topo_rules[rule.topology_type] = TopologyRule(rule.topology_type, rule.joint_type, rule.max_distance, **rule.kwargs)
        return direct_rules + category_rules + [rule for rule in topo_rules.values() if rule is not None]

    def apply_rules_to_model(self, model, handled_pairs=None):
        """Adds joints to model based on the given rules and elements.

        Parameters
        ----------
        model : :class:`~compas_timber.model.TimberModel`
            The timber model to which the rules will be applied.
        handled_pairs : list(set(:class:`~compas_timber.elements.TimberElement`, :class:`~compas_timber.elements.TimberElement`)), optional
            A list of already handled element pairs.

        Returns
        -------
        tuple(list(:class:`~compas_timber.connections.BeamJoiningError`), list(:class:`~compas_timber.connections.Cluster`), )
            A tuple containing a list of joining errors and a list of unjoined Clusters.

        """
        handled_pairs = handled_pairs or []
        max_rule_distance = max([rule.max_distance for rule in self.rules if rule.max_distance] + [self.max_distance])
        clusters = get_clusters_from_model(model, max_distance=max_rule_distance)
        clusters = self._remove_handled_pairs(clusters, handled_pairs)
        unjoined_clusters = self._joints_from_rules_and_clusters(model, self.rules, clusters, max_distance=self.max_distance)
        return self.joining_errors, unjoined_clusters

    def _remove_handled_pairs(self, clusters, handled_pairs):
        """Removes clusters from the list that have been handled."""
        remaining_clusters = []
        for cluster in clusters:
            if set(cluster.elements) not in handled_pairs:
                remaining_clusters.append(cluster)
        return remaining_clusters

    def _joints_from_rules_and_clusters(self, model, rules, clusters, max_distance=None):
        """Processes the JointRules and creates joints based on the clusters."""
        remaining_clusters = []
        for cluster in clusters:
            promoted = False
            for rule in rules:
                joint, error = rule.try_create_joint(model, cluster, max_distance=max_distance)
                if joint:
                    promoted = True
                    break
                if error:
                    self.joining_errors.append(error)  # should only happen with direct rules
                    break
            if not promoted:
                if len(cluster.joints) > 1:
                    sub_clusters = [Cluster([j]) for j in cluster.joints]
                    sub_remaining_clusters = self._joints_from_rules_and_clusters(model, rules, sub_clusters, max_distance=max_distance)
                    remaining_clusters.extend(sub_remaining_clusters)
                else:
                    remaining_clusters.append(cluster)
        return remaining_clusters


class JointRule(object):
    """Represents a rule for creating joints between timber elements.
    Parameters
    ----------
    joint_type : :class:`~compas_timber.joints.JointType`
        The type of joint to create.
    max_distance : float, optional
        The maximum distance to consider when creating joints. This will override the global max_distance.
    **kwargs : dict, optional
        Additional keyword arguments to pass to the joint creation method.
    """

    def __init__(self, joint_type, max_distance=None, **kwargs):
        self.joint_type = joint_type
        self.max_distance = max_distance
        self.kwargs = kwargs

    def _comply_topology(self, cluster, raise_error=False):
        """Checks if the given elements comply with the given topology.

        Parameters
        ----------
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster of elements to check.
        raise_error : bool, optional
            Whether to raise an error if the elements do not comply. If False, method will return False without raising an error.

        Returns
        -------
        bool
            True if the elements comply with the topology, False otherwise.

        """
        supported_topology = self.joint_type.SUPPORTED_TOPOLOGY
        if cluster.topology != supported_topology:
            if raise_error:
                raise BeamJoiningError(
                    beams=cluster.elements,
                    joint=self.joint_type,
                    debug_info="The cluster topology must be: {} for {}.".format(JointTopology.get_name(supported_topology), self.joint_type.__name__),
                    debug_geometries=[e.shape for e in cluster.elements],
                )
            return False
        return True

    def _comply_distance(self, cluster, max_distance=None, raise_error=False):
        """Checks if the distance between the elements in the cluster complies with the maximum distance.

        Parameters
        ----------
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster of elements to check.
        max_distance : float, optional
            The maximum distance to consider two elements as intersecting. This is the model max_distance and will be overridden if the rule has a max_distance.
        raise_error : bool, optional
            Whether to raise an error if the elements do not comply. If False, method will return False without raising an error.

        Returns
        -------
        bool
            True if the distance complies, False otherwise.
        """
        if not max_distance:
            return True

        distance = max([j.distance for j in cluster.joints])
        if distance > max_distance:
            if raise_error:
                raise BeamJoiningError(
                    beams=cluster.elements,
                    joint=self.joint_type,
                    debug_info="The distance between the elements is greater than the maximum allowed distance of {}.".format(max_distance),
                    debug_geometries=[e.shape for e in cluster.elements],
                )
            return False
        return True

    def _comply_element_count(self, cluster, raise_error=False):
        """Checks if the number of elements in the cluster complies with the joint's requirements.
        Parameters
        ----------
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster of elements to check.
        raise_error : bool, optional
            Whether to raise an error if the elements do not comply. If False, method will return False without raising an error.

        Returns
        -------
        bool
            True if the number of elements complies, False otherwise.
        """
        if not self.joint_type.element_count_complies(cluster.elements):
            if raise_error:
                raise BeamJoiningError(
                    beams=cluster.elements,
                    joint=self.joint_type,
                    debug_info="The number of elements in the cluster does not comply with the requirements for this joint.",
                    debug_geometries=[e.shape for e in cluster.elements],
                )
            return False
        return True


class DirectRule(JointRule):
    """Creates a Joint Rule that directly joins multiple elements. This essentially replaces the JointDefinition.

    Parameters
    ----------
    joint_type : cls(:class:`~compas_timber.connections.Joint`)
        The joint type to be applied to the elements.
    elements : list(:class:`~compas_timber.elements.TimberElement`)
        The elements to be joined.
    max_distance : float, optional
        The maximum distance to consider two elements as intersecting. This will override the global max_distance.
    kwargs : dict
        The keyword arguments to be passed to the joint.
    """

    def __init__(self, joint_type, elements, max_distance=None, **kwargs):
        super(DirectRule, self).__init__(joint_type, max_distance=max_distance, **kwargs)
        self.elements = elements

    def ToString(self):
        # GH doesn't know
        return repr(self)

    def __repr__(self):
        return "{}({}, {})".format(DirectRule, self.elements, self.joint_type)

    def _matches_cluster(self, cluster):
        """Checks if the elements in this rule match the elements in the cluster.
        Parameters
        ----------
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster to check against.

        Returns
        -------
        bool
            True if the elements match, False otherwise.
        """
        return set(self.elements) == cluster.elements

    def _comply_element_order(self, cluster, raise_error=False):
        """Checks if the element order in this rule complies with the element order in the cluster.
        This check is only performed for JointTopology.TOPO_T or JointTopology.TOPO_EDGE_FACE, where the topology requires a specific element order.
        Parameters
        ----------
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster to check against.
        raise_error : bool, optional
            Whether to raise an error if the element order does not comply. If False, method will return False without raising an error.

        Returns
        -------
        bool
            True if the element order complies, False otherwise.
        """
        if cluster.topology == JointTopology.TOPO_T or cluster.topology == JointTopology.TOPO_EDGE_FACE:
            if self.elements[0] != cluster.joints[0].elements[0]:
                if not raise_error:
                    return False
                raise BeamJoiningError(
                    beams=self.elements,
                    joint=self.joint_type,
                    debug_info="Beam roles must be reversed for topology of {}, try swapping elements in joint".format(self.joint_type.__name__),
                    debug_geometries=[e.shape for e in self.elements],
                )
        return True

    def try_create_joint(self, model, cluster, max_distance=None):
        """Returns a Joint if the given cluster's elements comply with this DirectRule.

        Parameters
        ----------
        model : :class:`~compas_timber.model.TimberModel`
            The model to which the elements and this joint belong.
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster of elements attemped to be joined.
        max_distance : float, optional
            The maximum distance to consider two elements as intersecting. Defaults to TOL.absolute.
            This is the general model.max_distance and will be overriden if the rule has a max_distance.

        Returns
        -------
        :class:`~compas_timber.connections.Joint`
            The joint created from the elements and the joint type.
        :class:`~compas_timber.errors.BeamJoiningError` or None
            Error that occurred during joint creation.

        """
        max_distance = self.max_distance or max_distance or TOL.absolute
        joint = None
        error = None
        if self._matches_cluster(cluster):
            try:
                self._comply_element_count(cluster, raise_error=True)
                self._comply_topology(cluster, raise_error=True)
                self._comply_element_order(cluster, raise_error=True)
                self._comply_distance(cluster, raise_error=True, max_distance=max_distance)
                self.joint_type.check_elements_compatibility(self.elements, raise_error=True)
                joint = self.joint_type.promote_cluster(model, cluster, reordered_elements=self.elements, **self.kwargs)
            except BeamJoiningError as bje:
                error = bje
        return joint, error


class CategoryRule(JointRule):
    """Based on the category attribute attached to the elements, this rule assigns

    Parameters
    ----------
    joint_type : cls(:class:`~compas_timber.connections.Joint`)
        The joint type to be applied to the elements.
    category_a : str
        The category of the first element.
    category_b : str
        The category of the second element.
    topos : list(:class:`~compas_timber.connections.JointTopology`), optional
        The topologies that are supported by this rule.
        TODO: reconsider how to implement this. was meant to filter/override joint application based on topology.
    max_distance : float, optional
        The maximum distance to consider two elements as intersecting.
    kwargs : dict
        The keyword arguments to be passed to the joint.
    """

    def __init__(self, joint_type, category_a, category_b, topos=None, max_distance=None, **kwargs):
        super(CategoryRule, self).__init__(joint_type, max_distance=max_distance, **kwargs)
        self.category_a = category_a
        self.category_b = category_b
        self.topos = topos or []

    def ToString(self):
        return repr(self)

    def __repr__(self):
        return "{}({}, {}, {}, {})".format(CategoryRule.__name__, self.joint_type.__name__, self.category_a, self.category_b, self.topos)

    def _comply_category_order(self, cluster, raise_error=False):
        """Checks if the order of the elements in the cluster complies with the rule's categories.
                Parameters
        ----------
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster of elements to check.
        raise_error : bool, optional
            Whether to raise an error if the elements do not comply. If False, method will return False without raising an error.

        Returns
        -------
        bool
            True if the order complies, False otherwise.
        """
        if cluster.topology == JointTopology.TOPO_T or cluster.topology == JointTopology.TOPO_EDGE_FACE:
            if cluster.joints[0].elements[0].attributes.get("category", None) != self.category_a:
                if not raise_error:
                    return False
                raise BeamJoiningError(
                    beams=self.elements,
                    joint=self.joint_type,
                    debug_info="Category roles must be reversed for topology of {}, try swapping categories in joint".format(self.joint_type.__name__),
                    debug_geometries=[e.shape for e in self.elements],
                )
        return True

    def _comply_categories(self, cluster):
        """Checks if the categories of the elements in the cluster are the same as the rule's categories.
        Parameters
        ----------
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster of elements to check.

        Returns
        -------
        bool
            True if the categories comply, False otherwise.
        """
        return set([e.attributes.get("category", None) for e in cluster.elements]) == {self.category_a, self.category_b}

    def _get_ordered_elements(self, cluster):
        """Returns the elements in the order of the rule's categories.
        This assumes self._comply_categories and self._comply_category_order have been checked and passed.

        Parameters
        ----------
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster of elements to order.

        Returns
        -------
        list(:class:`~compas_timber.elements.TimberElement`)
            The ordered list of elements.
        """
        elements = list(cluster.joints[0].elements)
        if cluster.joints[0].elements[0].attributes.get("category", None) != self.category_a:
            elements.reverse()
        return elements

    def try_create_joint(self, model, cluster, max_distance=None):
        """Returns a Joint if the given cluster's elements comply with this CategoryRule.

        Parameters
        ----------
        model : :class:`~compas_timber.model.TimberModel`
            The model to which the elements and this joint belong.
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster of elements to create the joint from.
        max_distance : float, optional
            The maximum distance to consider two elements as intersecting. Defaults to TOL.absolute.
            This is the general model.max_distance and will be overriden if the rule has a max_distance.

        Returns
        -------
        :class:`~compas_timber.connections.Joint` or None
            The joint created from the elements if the elements comply with the rule,
        :class:`~compas_timber.errors.BeamJoiningError` or None
            The error raised if the elements do not comply with the rule.

        """
        max_distance = self.max_distance or max_distance or TOL.absolute

        joint = None
        error = None
        if not self._comply_categories(cluster):
            return None, None
        if not self._comply_element_count(cluster):
            return None, None
        if not self._comply_topology(cluster):
            return None, None
        if not self._comply_category_order(cluster):
            return None, None
        if not self._comply_distance(cluster, max_distance=max_distance):
            return None, None
        elements = self._get_ordered_elements(cluster)
        if not self.joint_type.check_elements_compatibility(elements):
            return None, None
        try:
            joint = self.joint_type.promote_cluster(model, cluster, reordered_elements=elements, **self.kwargs)
        except BeamJoiningError as bje:
            error = bje
        return joint, error


class TopologyRule(JointRule):
    """for a given connection topology type (L,T,X,I,K...), this rule assigns a joint type.

    parameters
    ----------
    topology_type : constant(compas_timber.connections.JointTopology)
        The topology type to which the rule is applied.
    joint_type : cls(:class:`compas_timber.connections.Joint`)
        The joint type to be applied to this topology.
    max_distance : float, optional
        The maximum distance to consider two elements as intersecting.
        This will override a global max_distance if set.
    kwargs : dict
        The keyword arguments to be passed to the joint.
    """

    def __init__(self, topology_type, joint_type, max_distance=None, **kwargs):
        super(TopologyRule, self).__init__(joint_type, max_distance=max_distance, **kwargs)
        self.topology_type = topology_type

    def ToString(self):
        # GH doesn't know
        return repr(self)

    def __repr__(self):
        return "{}({}, {})".format(
            TopologyRule,
            self.topology_type,
            self.joint_type,
        )

    def try_create_joint(self, model, cluster, max_distance=None):
        """Returns a Joint if the given cluster's elements comply with this TopologyRule.
        It checks:
        that the max_distance is not exceeded,
        that the joint supports the topology of the elements.

        Parameters
        ----------
        model : :class:`~compas_timber.models.TimberModel`
            The timber model containing the elements.
        cluster : :class:`~compas_timber.connections.Cluster`
            The cluster of elements to create the joint from.
        max_distance : float, optional
            The maximum distance to consider two elements as intersecting. Defaults to TOL.absolute.
            This is the general model.max_distance and will be overriden if the rule has a max_distance.

        Returns
        -------
        :class:`~compas_timber.connections.Joint` or None
            The joint created from the elements if the elements comply with the rule,
        :class:`~compas_timber.errors.BeamJoiningError` or None
            The error raised if the elements do not comply with the rule.
        """
        max_distance = self.max_distance or max_distance or TOL.absolute

        joint = None
        error = None
        if not self._comply_element_count(cluster):
            return None, None
        if not self._comply_topology(cluster):
            return None, None
        if not self._comply_distance(cluster, max_distance=max_distance):
            return None, None
        if not self.joint_type.check_elements_compatibility(list(cluster.elements)):
            return None, None
        try:
            joint = self.joint_type.promote_cluster(model, cluster, **self.kwargs)
        except BeamJoiningError as bje:
            error = bje
        return joint, error


class Attribute:
    def __init__(self, attr_name, attr_value):
        self.name = attr_name
        self.value = attr_value

    def __str__(self):
        return "Attribute %s: %s" % (self.name, self.value)


def guess_joint_topology_2beams(beamA, beamB, tol=1e-6, max_distance=1e-6):
    # TODO: replace default max_distance ~ zero with global project precision

    [pa, ta], [pb, tb] = intersection_line_line_param(beamA.centerline, beamB.centerline, max_distance, True, tol)

    if ta is None or tb is None:
        # lines do not intersect within max distance or they are parallel
        return [None, None]

    def is_near_end(t, tol=tol):
        if abs(t) < tol:
            return True  # almost zero
        if abs(1.0 - t) < tol:
            return True  # almost 1
        return False

    xa = is_near_end(ta)
    xb = is_near_end(tb)

    if all([xa, xb]):
        # L-joint (both meeting at ends) TODO: this could also be an I-joint (splice) -> will need to check for angle between beams
        return ["L", (beamA, beamB)]
    elif any([xa, xb]):
        # T-joint (one meeting with the end along the other)
        if xa:
            # A:main, B:cross
            return ["T", (beamA, beamB)]
        if xb:
            # B:main, A:cross
            return ["T", (beamB, beamA)]
    else:
        # X-joint (both meeting somewhere along the line)
        return ["X", (beamA, beamB)]


def set_default_joints(model, x_default="x-lap", t_default="t-butt", l_default="l-miter"):
    beams = model.beams
    n = len(beams)

    connectivity = {"L": [], "T": [], "X": []}

    # find what kind of joint topology it looks like based on centerlines
    for i in range(n - 1):
        for j in range(i + 1, n):
            jointtype, beams_pair = guess_joint_topology_2beams(beams[i], beams[j])
            if jointtype:
                connectivity[jointtype].append(beams_pair)

    # Apply default joint types depending on the auto-found connectivity type:

    for beamA, beamB in connectivity["T"]:
        TButtJoint(beamA, beamB, model)

    for beamA, beamB in connectivity["L"]:
        LMiterJoint(beamA, beamB, model)

    for beamA, beamB in connectivity["X"]:
        pass


def get_clusters_from_model(model, max_distance=None, max_cluster_size=16):
    """Analyzes the model to find clusters of beams and plates. This will create JointCandidates and PlateJointCandidates in the model.
    Parameters
    ----------
    model : :class:`~compas_timber.model.TimberModel`
        The TimberModel to analyze.
    max_distance : float | None
        The maximum distance to consider for clustering. If None, a default distance is used.

    Returns
    -------
    list[:class:`~compas_timber.connections.Cluster`]
        A list of clusters found in the model.
    """
    # too high a max_cluster_size will lead to combinatorial explosion, even when efficient algorithm is used.
    # see: https://github.com/gramaziokohler/compas_timber/pull/705
    if max_cluster_size > 16:
        raise ValueError(f"max_cluster_size should not be too high to avoid combinatorial explosion, got {max_cluster_size}")
    model.connect_adjacent_beams(max_distance=max_distance)  # ensure that the model is connected before analyzing
    model.connect_adjacent_plates(max_distance=max_distance)  # ensure that the model is connected before analyzing
    analyzer = MaxNCompositeAnalyzer(model, n=max_cluster_size, max_distance=max_distance)
    clusters = analyzer.find()
    return clusters


class DebugInfomation(object):
    """Container for debugging information allowing visual inspection of joint and features related errors.

    Attributes
    ----------
    feature_errors : list(:class:`~compas_timber.consumers.FeatureApplicationError`)
        List of errors that occured during the application of features.
    joint_errors : list(:class:`~compas_timber.connections.BeamJoiningError`)
        List of errors that occured during the joining of beams.

    See Also
    --------
    :class:`~compas_timber.consumers.FeatureApplicationError`
    :class:`~compas_timber.connections.BeamJoiningError`

    """

    def __init__(self):
        self.fastener_errors = []
        self.feature_errors = []
        self.joint_errors = []

    def __repr__(self):
        return "{}({} feature errors, {} joining errors)".format(DebugInfomation.__name__, len(self.feature_errors), len(self.joint_errors))

    def ToString(self):
        return repr(self)

    @property
    def has_errors(self):
        return self.feature_errors or self.joint_errors

    def add_fastener_error(self, error):
        if isinstance(error, list):
            self.fastener_errors.extend(error)
        else:
            self.fastener_errors.append(error)

    def add_feature_error(self, error):
        if isinstance(error, list):
            self.feature_errors.extend(error)
        else:
            self.feature_errors.append(error)

    def add_joint_error(self, error):
        if isinstance(error, list):
            self.joint_errors.extend(error)
        else:
            self.joint_errors.append(error)



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper
import System


class FindBeamByRhinoGuid(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, beams: System.Collections.Generic.List[object], guid: System.Collections.Generic.List[System.Guid]):
        if not (beams and guid):
            return

        if not isinstance(guid, list):
            guid = [guid]
        Guid = [str(g) for g in Guid]
        FoundBeam = []
        for beam in beams:
            if beam.attributes.get("rhino_guid", None) in Guid:
                FoundBeam.append(beam)

        if not FoundBeam:
            self.AddRuntimeMessage(Warning, "No beams found!")

        return FoundBeam



# r: timber_design>=0.1.0
# venv: td_migration
"""Creates a Beam from a LineCurve."""

# flake8: noqa
import Grasshopper
import Rhino
import System
from compas.geometry import Brep
from compas.scene import Scene

from timber_design.workflow import ContainerDefinition
from compas_timber.elements import Wall
from timber_design.ghpython.ghcomponent_helpers import list_input_valid_cpython


class WallComponent(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(
        self, surface: System.Collections.Generic.List[Rhino.Geometry.Brep], thickness: System.Collections.Generic.List[float], config_set: System.Collections.Generic.List[object]
    ):
        # minimum inputs required
        if not list_input_valid_cpython(ghenv, surface, "Surface") or not list_input_valid_cpython(ghenv, thickness, "Thickness"):
            return

        if not config_set:
            config_set = [None]

        scene = Scene()
        containers = []
        # check list lengths for consistency
        N = len(surface)
        if len(thickness) not in (1, N):
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Error, " In 'T' I need either one or the same number of inputs as the Crv parameter.")

        # duplicate data if None or single value
        if len(thickness) != N:
            thickness = [thickness[0] for _ in range(N)]

        if len(config_set) != N:
            config_set = [config_set[0] for _ in range(N)]

        for srf, t, c_s in zip(surface, thickness, config_set):
            wall = Wall.from_brep(Brep.from_native(srf), t)

            containers.append(ContainerDefinition(wall, c_s))
            scene.add(wall.geometry)

        geo = scene.draw()

        return containers, geo



# r: timber_design>=0.1.0
# venv: td_migration
"""Generates a feature from BTLx type and input geometry."""

# flake8: noqa
import inspect

import Grasshopper
import Rhino.Geometry as rg
import System
from compas.geometry import Line
from compas_rhino.conversions import brep_to_compas
from compas_rhino.conversions import plane_to_compas
from compas_rhino.conversions import polyline_to_compas

from compas_timber.fabrication import BTLxFromGeometryDefinition
from compas_timber.fabrication import BTLxProcessing
from timber_design.ghpython import get_leaf_subclasses
from timber_design.ghpython import manage_cpython_dynamic_params
from timber_design.ghpython import rename_cpython_gh_output
from timber_design.ghpython import warning
from timber_design.ghpython import error
from timber_design.ghpython import message

import rhinoscriptsyntax as rs


class BTLxFromGeometry(Grasshopper.Kernel.GH_ScriptInstance):
    def __init__(self):
        super(BTLxFromGeometry, self).__init__()
        self.classes = {}
        self.geometry_count = 0
        for cls in get_leaf_subclasses(BTLxProcessing):
            functions = inspect.getmembers(cls, predicate=inspect.ismethod)
            if "from_shapes_and_element" in [func[0] for func in functions]:
                self.classes[cls.__name__] = cls

        if ghenv.Component.Params.Output[0].NickName == "Features":
            self.processing_type = None
        else:
            self.processing_type = self.classes.get(ghenv.Component.Params.Output[0].NickName, None)

    @property
    def component(self):
        return ghenv.Component  # type: ignore

    def RunScript(self, elements: System.Collections.Generic.List[object], *args):
        if not self.processing_type:
            warning(self.component, "Select Process type from context menu (right click)")
            return None
        else:
            message(self.component, self.processing_type.__name__)
            for arg, arg_name in zip(args, self.arg_names()[0 : self.geometry_count]):
                if arg is None:
                    warning(self.component, f"Input parameter {arg_name} failed to collect data")
            geometries = []
            for geo, arg_name in zip(args, self.arg_names()[0 : self.geometry_count]):
                if isinstance(geo, System.Guid):
                    geo = rs.coercegeometry(geo)  # guid to geometry
                if isinstance(geo, rg.LineCurve):
                    geometries.append(Line(geo.PointAtStart, geo.PointAtEnd))
                elif isinstance(geo, rg.Plane):
                    geometries.append(plane_to_compas(geo))
                elif isinstance(geo, rg.PolylineCurve):
                    geometries.append(polyline_to_compas(geo.ToPolyline()))
                elif isinstance(geo, rg.Brep):
                    geometries.append(brep_to_compas(geo))
                else:
                    error(self.component, f"Input parameter {arg_name} collect unusable data")

            if not geometries:
                warning(self.component, "no valid geometry collected")
                return

            kwargs = {"geometries": geometries}
            if elements:
                kwargs["elements"] = list(elements)
            for key, val in zip(self.arg_names()[self.geometry_count :], args[self.geometry_count :]):
                if val is not None:
                    kwargs[key] = val
            return BTLxFromGeometryDefinition(self.processing_type, **kwargs)

    def arg_names(self):
        names = inspect.getfullargspec(self.processing_type.from_shapes_and_element)[0][1:]
        count = 0
        for name in names:
            if name == "element":
                break
            else:
                count += 1
        self.geometry_count = count
        return [name for name in names if name != "element"]

    def AppendAdditionalMenuItems(self, menu):
        for name in self.classes.keys():
            item = menu.Items.Add(name, None, self.on_item_click)
            if self.processing_type and name == self.processing_type.__name__:
                item.Checked = True

    def on_item_click(self, sender, event_info):
        self.processing_type = self.classes[str(sender)]
        rename_cpython_gh_output(self.processing_type.__name__, 0, ghenv)
        manage_cpython_dynamic_params(self.arg_names(), ghenv, rename_count=0, permanent_param_count=1)
        ghenv.Component.ExpireSolution(True)



# r: timber_design>=0.1.0
# venv: td_migration
"""Shows the names of the connection topology types."""

# flake8: noqa
import re
import Grasshopper
import System
from compas_rhino.conversions import point_to_rhino
from compas_timber.connections import ConnectionSolver
from compas_timber.connections import GenericJoint
from compas_timber.connections import MaxNCompositeAnalyzer

from compas_timber.connections import JointTopology
from compas_timber.utils import intersection_line_line_param
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class ShowTopologyTypes(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, model):
        self.pt = []
        self.txt = []

        if not item_input_valid_cpython(ghenv, model, "model"):
            return

        # TODO: can we cash the clusters in the model?
        analyzer = MaxNCompositeAnalyzer(model, n=len(list(model.elements())))
        clusters = analyzer.find()

        for cluster in clusters:
            self.pt.append(cluster.joints[0].location)
            self.txt.append(cluster.joints[0].topology)

    def DrawViewportWires(self, arg):
        if ghenv.Component.Locked:
            return
        col = System.Drawing.Color.FromArgb(0, 0, 255, 0)
        # https://developer.rhino3d.com/api/RhinoCommon/html/M_Rhino_Display_DisplayPipeline_Draw2dText_5.htm
        for p, t in zip(self.pt, self.txt):
            arg.Display.Draw2dText(t, col, p, True, 12, "Verdana")



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import inspect
from System.Windows.Forms import ToolStripSeparator

import Grasshopper

from compas_timber.connections import PlateJoint
from compas_timber.connections import JointTopology
from compas_timber.connections import PlateTButtJoint
from timber_design.workflow import TopologyRule
from timber_design.ghpython.ghcomponent_helpers import get_leaf_subclasses
from timber_design.ghpython.ghcomponent_helpers import manage_cpython_dynamic_params
from timber_design.ghpython.ghcomponent_helpers import rename_cpython_gh_output


class EdgeFaceTopologyPlateJointRule(Grasshopper.Kernel.GH_ScriptInstance):
    def __init__(self):
        super(EdgeFaceTopologyPlateJointRule, self).__init__()
        self.classes = {}
        for cls in get_leaf_subclasses(PlateJoint):
            if cls.SUPPORTED_TOPOLOGY == JointTopology.TOPO_EDGE_FACE:
                self.classes[cls.__name__] = cls
        self.joint_type = self.classes.get(ghenv.Component.Params.Output[0].NickName, None)

    def RunScript(self, *args):
        if not self.joint_type:
            ghenv.Component.Message = "Default: PlateTButtJoint"
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "PlateTButtJoint is default, change in context menu (right click)")
            return TopologyRule(JointTopology.TOPO_EDGE_FACE, PlateTButtJoint)
        else:
            ghenv.Component.Message = self.joint_type.__name__
            kwargs = {}
            for i, val in enumerate(args):
                if val is not None:
                    kwargs[self.arg_names()[i]] = val
            return TopologyRule(JointTopology.TOPO_EDGE_FACE, self.joint_type, **kwargs)

    def AppendAdditionalMenuItems(self, menu):
        for name in self.classes.keys():
            item = menu.Items.Add(name, None, self.on_item_click)
            if self.joint_type and name == self.joint_type.__name__:
                item.Checked = True
        menu.Items.Add(ToolStripSeparator())

    def arg_names(self):
        return ["max_distance"]

    def on_item_click(self, sender, event_info):
        self.joint_type = self.classes[str(sender)]
        rename_cpython_gh_output(self.joint_type.__name__, 0, ghenv)
        manage_cpython_dynamic_params(self.arg_names(), ghenv, rename_count=0, permanent_param_count=0)
        ghenv.Component.ExpireSolution(True)



# r: timber_design>=0.1.0
# venv: td_migration
import Grasshopper
import Rhino
import System
from compas.scene import SceneObject
from compas_timber.elements import PlateFastener


class PlateFastenerComponent(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(
        self, outline: Rhino.Geometry.Polyline, thickness: float, cutouts: System.Collections.Generic.List[Rhino.Geometry.Polyline], main_beam_interface, cross_beam_interface
    ):
        outline_points = [point for point in outline] if outline else None
        cutout_points = []
        if cutouts:
            for cutout in cutouts:
                cutout_points.append([pt for pt in cutout])
        if main_beam_interface and cross_beam_interface:
            interfaces = [main_beam_interface, cross_beam_interface]
        else:
            interfaces = []
        fastener = PlateFastener(
            outline=outline_points,
            thickness=thickness,
            interfaces=interfaces,
            cutouts=cutout_points,
        )
        shape = None
        if outline and thickness:
            shape = SceneObject(item=fastener.shape).draw()

        return fastener, shape



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import math
import random

import Grasshopper
import System
import Rhino
import rhinoscriptsyntax as rs
from compas_rhino.conversions import frame_to_rhino
from timber_design.ghpython import item_input_valid_cpython

from Rhino import Render
from Rhino.Geometry import Interval
from Rhino.Geometry import Plane


class BakeBoxMap(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, model, map_size: System.Collections.Generic.List[float], bake: bool):
        if map_size and len(map_size) != 3:
            ghenv.Component.AddRuntimeMessage(
                Grasshopper.Kernel.GH_RuntimeMessageLevel.Error, "Input parameter MapSize requires exactly three float values (scale factors in x,y,z directions)"
            )
            return

        unit_system = Rhino.RhinoDoc.ActiveDoc.ModelUnitSystem

        if map_size:
            dimx, dimy, dimz = map_size
        else:
            # for the pine 251 material bitmap, rotated
            if unit_system == Rhino.UnitSystem.Meters:
                dimx = 0.2
                dimy = 0.2
                dimz = 1.0
            if unit_system == Rhino.UnitSystem.Centimeters:
                dimx = 20
                dimy = 20
                dimz = 100
            if unit_system == Rhino.UnitSystem.Millimeters:
                dimx = 200
                dimy = 200
                dimz = 1000

        if not item_input_valid_cpython(ghenv, model, "Model"):
            return

        if not bake:
            return

        try:
            frames = [frame_to_rhino(b.frame) for b in model.beams]
            breps = [beam.geometry.native_brep for beam in model.beams]

            if frames and breps:
                rs.EnableRedraw(False)

                for brep, frame in zip(breps, frames):
                    guid = Rhino.RhinoDoc.ActiveDoc.Objects.Add(brep)
                    boxmap = self.create_box_map(frame, dimx, dimy, dimz)
                    Rhino.RhinoDoc.ActiveDoc.Objects.ModifyTextureMapping(guid, 1, boxmap)
        finally:
            rs.EnableRedraw(True)

    @staticmethod
    def create_box_map(pln, sx, sy, sz):
        """
        pln: frame of beam box, where x=main axis, y=width, z=height
        sx,sy,sz: box map size in x,y,z direction
        """

        v = pln.YAxis
        w = pln.ZAxis
        pt = pln.Origin

        # random deviation
        a = math.pi * 0.5
        randangle = (random.random() - 0.5) * a
        v.Rotate(randangle, pln.XAxis)

        b = math.pi * 0.01
        randangle = (random.random() - 0.5) * b
        w.Rotate(randangle, pln.XAxis)

        randpos = sx * random.random()
        pt += pln.XAxis * randpos

        # create box mapping
        mappingPln = Plane(pt, w, v)
        dx = Interval(-sx * 0.5, sx * 0.5)
        dy = Interval(-sy * 0.5, sy * 0.5)
        dz = Interval(-sz * 0.5, sz * 0.5)

        BoxMap = Render.TextureMapping.CreateBoxMapping(mappingPln, dx, dy, dz, False)

        return BoxMap



import math

from compas.geometry import Brep
from compas.geometry import Frame
from compas.geometry import Line
from compas.geometry import NurbsCurve
from compas.geometry import Point
from compas.geometry import Vector
from compas.geometry import angle_vectors_signed
from compas.geometry import closest_point_on_segment
from compas.geometry import cross_vectors
from compas.geometry import distance_point_point_sqrd
from compas.geometry import dot_vectors
from compas.geometry import intersection_line_segment
from compas.geometry import intersection_segment_segment
from compas.geometry import matrix_from_frame_to_frame
from compas.geometry import offset_line
from compas.tolerance import TOL
from compas_timber.connections import ConnectionSolver
from compas_timber.connections import InterfaceLocation
from compas_timber.connections import InterfaceRole
from compas_timber.connections import JointTopology
from compas_timber.connections import LButtJoint
from compas_timber.connections import TButtJoint
from compas_timber.elements import Beam
from compas_timber.elements import OpeningType
from compas_timber.elements import Plate
from compas_timber.elements.features import BrepSubtraction

from .workflow import CategoryRule
from .workflow import JointDefinition


class WallSelector(object):
    """Selects walls based on their attributes."""

    def __init__(self, wall_attr, attr_vlaue):
        self._wall_attr = wall_attr
        self._attr_value = attr_vlaue

    def select(self, wall):
        value = getattr(wall, self._wall_attr, None)
        if value is None:
            return False
        else:
            return value == self._attr_value


class AnyWallSelector(object):
    def select(self, _):
        return True


class WallToWallInterface(object):
    """

    Parameters
    ----------
    face : :class:`compas.geometry.PlanarSurface`
        Planar surface representing the portion of the wall which touches the other wall.
        The face's normal points towards the other wall.
    topology : :class:`compas_timber.connections.JointTopology`
        The topology of the joint between the two walls.

    """

    def __init__(self, face, topology):
        self.face = face
        self.topology = topology


class BeamDefinition(object):
    """
    TODO: Move this to a separate module (potentially into workflow.py)

    Container for Beam attributes before beam is instantiated.

    Parameters
    ----------
    centerline : :class:`compas.geometry.Line`
        The centerline of the beam.
    width : float, optional
        The width of the beam.
    height : float, optional
        The height of the beam.
    z_axis : :class:`compas.geometry.Vector`, optional
        The z axis of the beam.
    normal : :class:`compas.geometry.Vector`, optional
        The normal of the beam.
    type : str, optional
        The type of the beam.
    polyline : :class:`compas.geometry.Polyline`, optional
        The polyline of the beam.
    parent : :class:`compas_timber.model.SurfaceAssembly` or :class:`compas_timber.model.SurfaceAssembly.Window`
        The parent of the beam.

    Attributes
    ----------
    centerline : :class:`compas.geometry.Line`
        The centerline of the beam element.
    width : float
        The width of the beam element.
    height : float
        The height of the beam element.
    z_axis : :class:`compas.geometry.Vector`
        The z axis of the parent (Not the Beam).
    normal : :class:`compas.geometry.Vector`
        The normal of the parent.


    """

    def __init__(
        self,
        centerline,
        width=None,
        height=None,
        z_axis=None,  # TODO: remove
        normal=None,
        type=None,
        parent=None,
    ):
        self.original_centerline = centerline
        self.centerline = Line(centerline[0], centerline[1])
        self._width = width
        self._height = height
        self._normal = normal
        self.type = type
        self.parent = parent  # TODO: remove

    def __repr__(self):
        return "BeamDefinition({}, {}, {}, {})".format(self.centerline, self.width, self.height, self.type)

    @property
    def width(self):
        # TODO: provide this upon creation
        return self._width if self._width else self.parent.beam_dimensions[self.type][0]

    @property
    def height(self):
        # TODO: provide this upon creation
        return self._height if self._height else self.parent.beam_dimensions[self.type][1]

    @property
    def z_aligned_centerline(self):
        # TODO: provide this upon creation
        if dot_vectors(self.centerline.direction, self.parent.z_axis) < 0:
            return Line(self.centerline.end, self.centerline.start)
        else:
            return self.centerline

    @property
    def normal(self):
        # TODO: provide this upon creation
        return self._normal if self._normal else self.parent.normal

    def offset(self, distance, reverse_dir=False):
        normal = self.normal if not reverse_dir else -self.normal
        line = offset_line(self.centerline, distance, normal)
        self.centerline = Line(line[0], line[1])

    def translate(self, vector):
        self.centerline.transform(vector)

    def to_beam(self):
        # TODO: this is quite stiff, if I know the corner, I can just create beam using Beam(...)
        centerline = Line(*self.centerline)
        centerline.translate(self.normal * 0.5 * self.height)
        beam = Beam.from_centerline(centerline, self.width, self.height, self.normal)
        beam.name = "{}_{}".format(self.type, beam.guid)
        beam.attributes["category"] = self.type
        return beam

    def set_centerline(self, line):
        self.centerline = line


class Window(object):
    """

    # TODO: is this an Element maybe?

    A window object for the SurfaceAssembly.

    Parameters
    ----------
    outline : :class:`compas.geometry.Polyline`
        The outline of the window.
    sill_height : float, optional
        The height of the sill.
    header_height : float, optional
        The height of the header.
    parent : :class:`compas_timber.model.SurfaceAssembly`
        The parent of the window.

    Attributes
    ----------
    outline : :class:`compas.geometry.Polyline`
        The outline of the window.
    sill_height : float
        The height of the sill.
    header_height : float
        The height of the header.
    parent : :class:`compas_timber.model.SurfaceAssembly`
        The parent of the window.
    z_axis : :class:`compas.geometry.Vector`
        The z axis of the parent.
    normal : :class:`compas.geometry.Vector`
        The normal of the parent.
    beam_dimensions : dict
        The beam dimensions of the parent.
    beam_definions : list of :class:`compas_timber.model.SurfaceAssembly.BeamDefinition`
        The beam_definions of the window.
    length : float
        The length of the window.
    height : float
        The height of the window.
    frame : :class:`compas.geometry.Frame`
        The frame of the window.

    """

    def __init__(self, outline, beam_dimensions, wall_frame, wall_thickness, tolerance, sheeting_inside=None, sheeting_outside=None, lintel_posts=None):
        self.beam_dimensions = beam_dimensions
        self._wall_frame = wall_frame
        self._sheeting_inside = sheeting_inside
        self._sheeting_outside = sheeting_outside
        self._lintel_posts = lintel_posts
        self._wall_thickness = wall_thickness
        self.outline = outline
        self.sill_height = self.beam_dimensions["sill"][0]
        self.header_height = self.beam_dimensions["header"][0]
        self.z_axis = self._wall_frame.yaxis
        self.normal = self._wall_frame.zaxis
        self._beam_definitions = []
        self._length = None
        self._height = None
        self._frame = None
        self.dist_tolerance = tolerance

    @property
    def jack_studs(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "jack_stud"]

    @property
    def sills(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "sill"]

    @property
    def headers(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "header"]

    @property
    def length(self):
        if not self._length:
            _ = self.frame
        return self._length

    @property
    def height(self):
        if not self._height:
            _ = self.frame
        return self._height

    @property
    def frame(self):
        # if not self._frame:
        #     self._frame, self._panel_length, self._panel_height = get_frame(self.points, self.parent.normal, self.zaxis)
        return self._frame

    @property
    def boolean_feature(self):
        offset = self._sheeting_inside if self._sheeting_inside else 0
        so = self._sheeting_outside if self._sheeting_outside else 0
        thickness = offset + so + self._wall_thickness

        crv = self.outline.copy()
        crv.translate(self.normal * -offset)

        vol = Brep.from_extrusion(NurbsCurve.from_points(crv.points, degree=1), self.normal * thickness)

        # negative volume will cause weird boolean result
        if vol.volume < 0:
            # TODO: remove once this is release in compas core
            if hasattr(vol, "flip"):
                vol.flip()

        return BrepSubtraction(vol)

    def create_elements(self):
        # ^  3 ---- 2
        # |  |      |
        # z  0 ---- 1
        #    x -->
        top_segment = self.outline.lines[2]
        bottom_segment = self.outline.lines[0]
        left_segment = self.outline.lines[3]
        right_segment = self.outline.lines[1]

        header = BeamDefinition(top_segment, parent=self, type="header")
        sill = BeamDefinition(bottom_segment, parent=self, type="sill")

        stud_roles = "jack_stud" if self._lintel_posts else "king_stud"
        studs = []
        studs.append(BeamDefinition(left_segment, parent=self, type=stud_roles))
        studs.append(BeamDefinition(right_segment, parent=self, type=stud_roles))

        beam_definitions = [header, sill] + studs

        offset_elements(beam_definitions, offset_inside=False)

        shorten_edges_to_fit_between_plates(studs, [header, sill])

        if self._lintel_posts:
            for beam_def in self.jack_studs:
                offset = (self.beam_dimensions["jack_stud"][0] + self.beam_dimensions["king_stud"][0]) / 2
                king_line = offset_line(beam_def.centerline, offset, self.normal)
                beam_definitions.append(BeamDefinition(king_line, type="king_stud", parent=self))
        return beam_definitions


class Door(Window):
    """TODO: revise when we know where this is going, maybe no need for classes here beyond Opening"""

    def __init__(self, outline, beam_dimensions, wall_frame, wall_thickness, tolerance, sheeting_inside=None, sheeting_outside=None, lintel_posts=None):
        super(Door, self).__init__(outline, beam_dimensions, wall_frame, wall_thickness, tolerance, sheeting_inside, sheeting_outside, lintel_posts)

    def create_elements(self):
        elements = super(Door, self).create_elements()
        return [e for e in elements if e.type != "sill"]


class WallPopulatorConfigurationSet(object):
    """Contains one or more configuration set for the WallPopulator.

    wall_selector can be used to apply different configurations to different walls based on e.g. their name.

    Parameters
    ----------
    stud_spacing : float
        Space between the studs.
    beam_width : float
        Width of the beams.
    tolerance : :class:`compas_tolerances.Tolerance`, optional
        The tolerance for the populator.
    sheeting_outside : float, optional
        The thickness of the sheeting outside.
    sheeting_inside : float, optional
        The thickness of the sheeting inside.
    lintel_posts : bool, optional
        Whether to use lintel posts.
    edge_stud_offset : float, optional
        Additional offset for the edge studs.
    custom_dimensions : dict, optional
        Custom cross section for the beams, by category. (e.g. {"king_stud": (120, 60)})
    joint_overrides : list(`compas_timber.workflow.CategoryRule), optional
        List of joint rules to override the default ones.
    connection_details : dict, optional
        Mapping of `JointTopology` to and instace of ConnectionDetail class.

    """

    def __init__(
        self,
        stud_spacing,
        beam_width,
        z_axis=None,
        tolerance=None,
        sheeting_outside=None,
        sheeting_inside=None,
        lintel_posts=True,
        edge_stud_offset=0.0,
        custom_dimensions=None,
        joint_overrides=None,
        wall_selector=None,
        connection_details=None,
    ):
        self.stud_spacing = stud_spacing
        self.beam_width = beam_width
        self.z_axis = z_axis or Vector.Zaxis()
        self.tolerance = tolerance or TOL
        self.sheeting_outside = sheeting_outside
        self.sheeting_inside = sheeting_inside
        self.lintel_posts = lintel_posts
        self.edge_stud_offset = edge_stud_offset or 0.0
        self.custom_dimensions = custom_dimensions
        self.joint_overrides = joint_overrides
        self.wall_selector = wall_selector or AnyWallSelector()
        self.connection_details = connection_details or {}

    def __str__(self):
        return "WallPopulatorConfigurationSet({}, {}, {})".format(self.stud_spacing, self.beam_width, self.z_axis)

    @classmethod
    def default(cls, stud_spacing, beam_width):
        return cls(stud_spacing, beam_width)


class WallPopulator(object):
    """Create a timber assembly from a surface.

    Parameters
    ----------
    configuration_set : :class:`WallPopulatorConfigurationSet`
        The configuration for this wall populator.
    wall : :class:`compas_timber.elements.Wall`
        The wall for this populater to fill with beams.
    interfaces : optional, list of :class:`WallToWallInterface`
        The interfaces of the wall.

    Attributes
    ----------
    beams : list of :class:`compas_timber.elements.Beam`
        The beams of the assembly.
    rules : list of :class:`compas_timber.design.CategoryRule`
        The rules for the assembly.
    centerlines : list of :class:`compas.geometry.Line`
        The centerlines of the beams.
    normal : :class:`compas.geometry.Vector`
        The normal of the surface.
    panel_length : float
        The length of the panel.
    panel_height : float
        The height of the panel.
    frame : :class:`compas.geometry.Frame`
        The frame of the assembly.
    jack_studs : list of :class:`BeamDefinition`
        The jack studs of the assembly.
    king_studs : list of :class:`BeamDefinition`
        The king studs of the assembly.
    edge_studs : list of :class:`BeamDefinition`
        The edge studs of the assembly.
    studs : list of :class:`BeamDefinition`
        The studs of the assembly.
    sills : list of :class:`BeamDefinition`
        The sills of the assembly.
    headers : list of :class:`BeamDefinition`
        The headers of the assembly.
    plates : list of :class:`BeamDefinition`

    """

    BEAM_CATEGORY_NAMES = ["stud", "king_stud", "jack_stud", "edge_stud", "plate", "header", "sill", "detail"]

    def __init__(self, configuration_set, wall, interfaces=None):
        self._wall = wall
        self._config_set = configuration_set
        self._z_axis = wall.frame.yaxis  # up
        self.normal = wall.frame.zaxis  # out
        self.outer_polyline = wall.outline
        self.inner_polylines = wall.openings
        self.edges = []
        self._elements = []
        self._beam_definitions = []
        self._rules = []
        self._features = []
        self.beam_dimensions = {}
        self.dist_tolerance = configuration_set.tolerance.relative

        self.frame = wall.frame
        self.panel_length = wall.length
        self.panel_height = wall.height

        self._interfaces = interfaces or []
        self._adjusted_segments = {}
        self._plate_segments = {}
        self._detail_obbs = []
        self._openings = []

        for key in self.BEAM_CATEGORY_NAMES:
            self.beam_dimensions[key] = (configuration_set.beam_width, wall.thickness)

        if self._config_set.custom_dimensions:
            dimensions = self._config_set.custom_dimensions
            for key, value in dimensions.items():
                if value:
                    self.beam_dimensions[key] = value

    def __repr__(self):
        return "WallPopulator({}, {})".format(self._config_set, self._wall)

    @property
    def z_axis(self):
        cross = cross_vectors(self.normal, self._z_axis)
        return Vector(*cross_vectors(cross, self.normal))

    @property
    def default_rules(self):
        edge_plate_joint = LButtJoint if TOL.is_zero(self._config_set.edge_stud_offset) else TButtJoint
        return [
            CategoryRule(edge_plate_joint, "edge_stud", "plate"),
            CategoryRule(TButtJoint, "detail", "plate"),  # TODO: have the details define this
            CategoryRule(LButtJoint, "detail_edge", "plate"),  # TODO: have the details define this
            CategoryRule(TButtJoint, "stud", "plate"),
            CategoryRule(TButtJoint, "stud", "header"),
            CategoryRule(TButtJoint, "stud", "sill"),
            CategoryRule(LButtJoint, "jack_stud", "plate"),
            CategoryRule(TButtJoint, "jack_stud", "plate"),
            CategoryRule(LButtJoint, "jack_stud", "header"),
            CategoryRule(TButtJoint, "jack_stud", "header"),
            CategoryRule(TButtJoint, "king_stud", "plate"),
            CategoryRule(LButtJoint, "king_stud", "plate"),
            CategoryRule(TButtJoint, "king_stud", "sill"),
            CategoryRule(TButtJoint, "king_stud", "header"),
            CategoryRule(TButtJoint, "sill", "jack_stud"),
        ]

    @property
    def rules(self):
        if not self._rules:
            self._rules = self.default_rules
            if self._config_set.joint_overrides:
                for rule in self._config_set.joint_overrides:
                    rule_set = set([rule.category_a, rule.category_b])
                    for i, _rule in enumerate(self._rules):
                        _set = set([_rule.category_a, _rule.category_b])
                        if rule_set == _set:
                            self._rules[i] = rule
                            break
        return self._rules

    @property
    def centerlines(self):
        return [beam_def.centerline for beam_def in self.beam_definitions]

    @property
    def elements(self):
        elements = []
        for beam_def in self._beam_definitions:
            try:
                elements.append(beam_def.to_beam())
            except Exception:
                print("Error creating beam from centerline: {}".format(beam_def.centerline))
                print("Beam role is: {}".format(beam_def.type))
        elements.extend(self._elements)
        return elements

    @property
    def features(self):
        return self._features

    @property
    def plate_elements(self):
        for plate in self._elements:
            if isinstance(plate, Plate):
                yield plate

    @property
    def points(self):
        return self.outer_polyline.points

    @property
    def jack_studs(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "jack_stud"]

    @property
    def king_studs(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "king_stud"]

    @property
    def edge_studs(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "edge_stud"]

    @property
    def studs(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "stud"]

    @property
    def sills(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "sill"]

    @property
    def headers(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "header"]

    @property
    def plates(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "plate"]

    @classmethod
    def beam_category_names(cls):
        return WallPopulator.BEAM_CATEGORY_NAMES

    @classmethod
    def from_model(cls, model, configuration_sets):
        # type: (TimberModel, List[WallPopulatorConfigurationSet]) -> List[WallPopulator]
        """matches configuration sets to walls and returns a list of WallPopulator instances, each per wall"""
        # TODO: make sure number of walls and configuration sets match
        walls = list(model.slabs)  # TODO: these are anoying, consider making these lists again
        if len(walls) != len(configuration_sets):
            raise ValueError("Number of walls and configuration sets do not match")

        wall_populators = []
        for wall in walls:
            for config_set in configuration_sets:
                if config_set.wall_selector.select(wall):
                    interfaces = [interaction.get_interface_for_wall(wall) for interaction in model.get_interactions_for_element(wall)]
                    wall_populators.append(cls(config_set, wall, interfaces))
                    break
        return wall_populators

    def create_elements(self):
        """Does the actual populating of the wall

        creates and returns all the elements in the wall, returns also the joint definitions
        """
        self.generate_perimeter_beams()
        self.generate_openings()
        self.generate_studs()
        self.generate_plates()
        elements = self.elements
        return elements

    def create_joint_definitions(self, elements, max_distance=None):
        beams = [element for element in elements if element.is_beam]
        solver = ConnectionSolver()
        found_pairs = solver.find_intersecting_pairs(beams, rtree=True, max_distance=self.dist_tolerance)

        joint_definitions = []
        max_distance = max_distance or 0.0
        max_distance = max(self._config_set.beam_width, max_distance)  # oterwise L's become X's
        for pair in found_pairs:
            beam_a, beam_b = pair
            detected_topo, beam_a, beam_b = solver.find_topology(beam_a, beam_b, max_distance=max_distance)
            if detected_topo == JointTopology.TOPO_UNKNOWN:
                continue

            for rule in self.rules:
                if rule.comply(pair, model_max_distance=max_distance) and rule.joint_type.SUPPORTED_TOPOLOGY == detected_topo:
                    if rule.joint_type == LButtJoint:
                        beam_a, beam_b = rule.reorder([beam_a, beam_b])
                    joint_definitions.append(JointDefinition(rule.joint_type, [beam_a, beam_b], **rule.kwargs))
                    # break # ?
        return joint_definitions

    def generate_perimeter_beams(self):
        # for each interface, find the appropriate connection detail (depending on the topology)
        # first the interfaces are handled, then the remaining sides are handled by the default connection details
        handled_sides = set()

        # TODO: move to Wall
        # add any remaining sides if have not been handled by any of the connection details
        # ^  3 ---- 2
        # |  |      |
        # z  0 ---- 1
        #    x -->
        front_segment = Line(self.points[1], self.points[2])
        back_segment = Line(self.points[3], self.points[0])
        top_segment = Line(self.points[2], self.points[3])
        bottom_segment = Line(self.points[0], self.points[1])
        self._adjusted_segments = {"front": front_segment, "back": back_segment, "top": top_segment, "bottom": bottom_segment}

        perimeter_beams = []
        for interface in self._interfaces:
            connection_detail = self._config_set.connection_details.get(interface.topology, None)

            if connection_detail:
                if interface.interface_role == InterfaceRole.MAIN:
                    perimeter_beams.extend(connection_detail.create_elements_main(interface, self._wall, self._config_set))
                    connection_detail.adjust_segments_main(interface, self._wall, self._config_set, self._adjusted_segments)
                elif interface.interface_role == InterfaceRole.CROSS:
                    connection_detail.adjust_segments_cross(interface, self._wall, self._config_set, self._adjusted_segments)
                    perimeter_beams.extend(connection_detail.create_elements_cross(interface, self._wall, self._config_set))

                handled_sides.add(interface.interface_type)

        top_segment = self._adjusted_segments["top"]
        bottom_segment = self._adjusted_segments["bottom"]
        front_segment = self._adjusted_segments["front"]
        back_segment = self._adjusted_segments["back"]

        if InterfaceLocation.FRONT not in handled_sides:
            perimeter_beams.append(BeamDefinition(front_segment, parent=self, type="edge_stud"))

        if InterfaceLocation.BACK not in handled_sides:
            perimeter_beams.append(BeamDefinition(back_segment, parent=self, type="edge_stud"))

        if InterfaceLocation.TOP not in handled_sides:
            perimeter_beams.append(BeamDefinition(top_segment, parent=self, type="plate"))

        if InterfaceLocation.BOTTOM not in handled_sides:
            perimeter_beams.append(BeamDefinition(bottom_segment, parent=self, type="plate"))

        edge_studs = [beam_def for beam_def in perimeter_beams if beam_def.type == "edge_stud"]
        plate_beams = [beam_def for beam_def in perimeter_beams if beam_def.type == "plate"]

        if self._wall.is_wall:
            offset_elements(edge_studs + plate_beams)
        else:
            # HACK alert! slabs seem to want it differently, get to the bottom of this
            if self.normal.z < 0:
                offset_elements(edge_studs + plate_beams, offset_inside=False)
            else:
                offset_elements(edge_studs + plate_beams)

        shorten_edges_to_fit_between_plates(edge_studs, plate_beams, self.dist_tolerance)

        self._beam_definitions.extend(perimeter_beams)

        # TODO: handle lintel posts

    def get_interior_segment_indices(self, polyline):
        points = polyline.points[0:-1]
        out = []
        for index in range(len(points)):
            if index == 0:
                angle = angle_vectors_signed(points[-1] - points[0], points[1] - points[0], self.normal, deg=True)
            elif index == len(points) - 1:
                angle = angle_vectors_signed(points[-2] - points[-1], points[0] - points[-1], self.normal, deg=True)
            else:
                angle = angle_vectors_signed(points[index - 1] - points[index], points[index + 1] - points[index], self.normal, deg=True)
            if angle > 0:
                out.append(index)
        if len(out) > 0:
            out.insert(0, out[0] - 1)
        return set(out)

    def generate_openings(self):
        for opening in self.inner_polylines:
            if opening.opening_type == OpeningType.DOOR:
                element = Door(
                    opening.polyline,
                    self.beam_dimensions,
                    self._wall.frame,
                    self._wall.thickness,
                    self.dist_tolerance,
                    self._config_set.sheeting_inside,
                    self._config_set.sheeting_outside,
                    # self._config_set.lintel_posts,
                )
            else:
                element = Window(
                    opening.polyline,
                    self.beam_dimensions,
                    self._wall.frame,
                    self._wall.thickness,
                    self.dist_tolerance,
                    self._config_set.sheeting_inside,
                    self._config_set.sheeting_outside,
                    # self._config_set.lintel_posts,
                )

            self._beam_definitions.extend(element.create_elements())
            self._openings.append(element)

    def generate_studs(self):
        self.generate_stud_lines()
        self.trim_jack_studs()
        self.trim_king_studs()
        self.trim_studs()
        self.cull_overlaps()

    def generate_stud_lines(self):
        x_position = self._config_set.stud_spacing
        while x_position < self.panel_length - self._config_set.beam_width:
            start_point = Point(x_position, 0, 0)
            start_point.transform(matrix_from_frame_to_frame(Frame.worldXY(), self.frame))
            line = Line.from_point_and_vector(start_point, self.z_axis * self.panel_height)
            self._beam_definitions.append(BeamDefinition(line, type="stud", parent=self))
            x_position += self._config_set.stud_spacing

    def get_beam_intersections(self, beam_def, *element_lists_to_intersect):
        intersections = []
        dots = []
        for element_list in element_lists_to_intersect:
            for element_to_intersect in element_list:
                point = intersection_line_segment(beam_def.z_aligned_centerline, element_to_intersect.centerline, 0.01)[0]
                if point:
                    intersections.append(point)
        if len(intersections) > 1:
            intersections.sort(key=lambda x: dot_vectors(x, self.z_axis))
            dots = [dot_vectors(Vector.from_start_end(beam_def.z_aligned_centerline.start, x), self.z_axis) / beam_def.centerline.length for x in intersections]
        return intersections, dots

    def trim_jack_studs(self):
        for beam_def in self.jack_studs:
            intersections, dots = self.get_beam_intersections(beam_def, self.plates, self.headers)
            if len(intersections) > 1:
                bottom = None
                for i, dot in enumerate(dots):
                    if dot < 0:
                        bottom = intersections[i]
                top = beam_def.z_aligned_centerline.end
                if not bottom:
                    bottom = beam_def.z_aligned_centerline.start
            beam_def.set_centerline(Line(bottom, top))

    def trim_king_studs(self):
        for beam_def in self.king_studs:
            intersections, dots = self.get_beam_intersections(beam_def, self.plates, self.headers, self.sills)
            if len(intersections) > 1:
                bottom, top = None, None
                for i, dot in enumerate(dots):
                    if dot < -0.01:
                        bottom = intersections[i]  # last intersection below sill
                    if dot > 1.01:
                        top = intersections[i]  # first intersection above header
                        break
                if not bottom:
                    bottom = beam_def.z_aligned_centerline.start
                if not top:
                    top = beam_def.z_aligned_centerline.end
                beam_def.set_centerline(Line(bottom, top))

    def trim_studs(self):
        stud_elements = []
        while len(self.studs) > 0:
            for beam_def in self._beam_definitions:
                if beam_def.type == "stud":
                    intersections, _ = self.get_beam_intersections(beam_def, self.plates, self.headers, self.sills)
                    while len(intersections) > 1:
                        top = intersections.pop()
                        bottom = intersections.pop()
                        stud_elements.append(BeamDefinition(Line(bottom, top), type="stud", parent=self))
                    self._beam_definitions.remove(beam_def)
        self._beam_definitions.extend(stud_elements)

    def cull_overlaps(self):
        studs = self.studs
        for beam_def in studs:
            for other_element in self.king_studs + self.jack_studs + self.edge_studs:
                if self.distance_between_elements(beam_def, other_element) < (self.beam_dimensions[beam_def.type][0] + self.beam_dimensions[other_element.type][0]) / 2:
                    self._beam_definitions.remove(beam_def)
                    break

        # removed studs that are inside details
        for beam_def in studs:
            for interface in self._interfaces:
                detail = self._config_set.connection_details.get(interface.topology, None)
                if not detail:
                    continue

                if interface.interface_role == InterfaceRole.MAIN:
                    detail_obb = detail.get_detail_obb_main(interface, self._config_set, self._wall)
                else:
                    detail_obb = detail.get_detail_obb_cross(interface, self._config_set, self._wall)

                if detail_obb.contains_point(beam_def.centerline.midpoint):
                    self._beam_definitions.remove(beam_def)
                    break

    def distance_between_elements(self, element_one, element_two):
        distances = []
        for pt in element_one.centerline:
            cp = closest_point_on_segment(pt, element_two.centerline)
            distances.append(distance_point_point_sqrd(pt, cp))
        for pt in element_two.centerline:
            cp = closest_point_on_segment(pt, element_one.centerline)
            distances.append(distance_point_point_sqrd(pt, cp))
        return math.sqrt(min(distances))

    def generate_plates(self):
        plates = []
        if self._config_set.sheeting_inside:
            pline = self.outer_polyline.translated(self.frame.zaxis * (-self._config_set.sheeting_inside))
            plates.append(Plate.from_outline_thickness(pline, self._config_set.sheeting_inside))
        if self._config_set.sheeting_outside:
            pline = self.outer_polyline.translated(self.frame.zaxis * (self._wall.thickness))
            plates.append(Plate.from_outline_thickness(pline, self._config_set.sheeting_outside))
            # this assumes that self.frame.zaxis points towards the outside of the wall
        for opening in self._openings:
            for plate in plates:
                plate.add_feature(opening.boolean_feature)

        self._elements.extend(plates)


def shorten_edges_to_fit_between_plates(beams_to_fit, beams_to_fit_between, dist_tolerance=None):
    """Shorten the beams to fit between the plates. The given beam definitions are modified in-place.

    Parameters
    ----------
    beams_to_fit : list of :class:`BeamDefinition`
        The beams to fit between the plates.
    beams_to_fit_between : list of :class:`BeamDefinition`
        The plates to fit between.
    dist_tolerance : float, optional
        The distance tolerance for the intersection check.
        Default is ``TOL.absolute``.

    """
    plate_a, plate_b = beams_to_fit_between  # one is top, one is bottom, might be important to distinguish at some point.
    dist_tolerance = dist_tolerance or TOL.absolute
    for stud in beams_to_fit:
        start_point, _ = intersection_segment_segment(stud.centerline, plate_a.centerline, dist_tolerance)
        end_point, _ = intersection_segment_segment(stud.centerline, plate_b.centerline, dist_tolerance)
        if start_point and end_point:
            stud.centerline = Line(start_point, end_point)


def offset_elements(element_loop, edge_stud_offset=None, offset_inside=True):
    """Offset elements towards the inside of the wall. The given beam definitions are modified in-place.

    Parameters
    ----------
    element_loop : list of :class:`BeamDefinition`
        The elements to offset.
    edge_stud_offset : float, optional
        The additional offset for edge studs towards the inside of the wall to account for bending.
        Default is ``0.0``.
    offset_inside : bool, optional
        Offset the elements towards the inside of the wall.
        Default is ``True``. If ``False``, the elements are offset towards the outside.

    """
    # TODO: rename to offset_perimeter_elements
    edge_studs = [beam_def for beam_def in element_loop if beam_def.type in "edge_stud"]
    edge_stud_offset = edge_stud_offset or 0.0
    for beam_def in element_loop:
        # polyline is at the center of the beam's face, push it towards the inside
        beam_def.offset(beam_def.width / 2, reverse_dir=not offset_inside)

    for stud in edge_studs:
        # configurable additional offset for edge studs towards the inside of the wall to account for bending
        stud.offset(edge_stud_offset)



try:
    import Grasshopper  # type: ignore
    import RhinoCodePluginGH.Parameters  # type: ignore
except (ImportError, SyntaxError):
    pass


def warning(component, message):
    """Add a warning message to the component.

    Parameters
    ----------
    component : Grasshopper.Kernel.IGH_Component
        The component instance. Pre-Rhino8 use `self`. Post-Rhino8 use `ghenv.Component`.
    message : str
        The message to display.
    """
    component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, message)


def error(component, message):
    """Add an error message to the component.

    Parameters
    ----------
    component : Grasshopper.Kernel.IGH_Component
        The component instance. Pre-Rhino8 use `self`. Post-Rhino8 use `ghenv.Component`.
    message : str
        The message to display.
    """
    component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Error, message)


def remark(component, message):
    """Add a remark message to the component.

    Parameters
    ----------
    component : Grasshopper.Kernel.IGH_Component
        The component instance. Pre-Rhino8 use `self`. Post-Rhino8 use `ghenv.Component`.
    message : str
        The message to display.
    """
    component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Remark, message)


def message(component, message):
    """Add a text that will appear under the component.

    Parameters
    ----------
    component : Grasshopper.Kernel.IGH_Component
        The component instance. Pre-Rhino8 use `self`. Post-Rhino8 use `ghenv.Component`.
    message : str
        The message to display.
    """
    component.Message = message


def list_input_valid(component, Param, name):
    if not Param:
        component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Input parameter %s failed to collect data" % name)
    else:
        if all([_ is None for _ in Param]):
            component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Input parameter %s failed to collect data" % name)
        elif any([_ is None for _ in Param]):
            component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Remark, "Input parameter %s contains some Null values" % name)
            return True
        else:
            return True
    return False


def item_input_valid(component, Param, name):
    if not Param:
        component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Input parameter %s failed to collect data" % name)
    else:
        return True
    return False


def get_leaf_subclasses(cls):
    subclasses = []
    for subclass in cls.__subclasses__():
        if not get_leaf_subclasses(subclass):
            subclasses.append(subclass)
        subclasses.extend(get_leaf_subclasses(subclass))
    return subclasses


def get_createable_joints():
    """Returns the list of creatable joint types.

    TODO: this is currently hardcoded but can be done duynamically, however, a better distinction should be made first between abstract and concrete joints.
    TOOD: a typical way of doing that would be to declare the non-creatable base classes abstract and filter those.
    """
    from compas_timber.connections import BallNodeJoint
    from compas_timber.connections import LButtJoint
    from compas_timber.connections import LFrenchRidgeLapJoint
    from compas_timber.connections import LLapJoint
    from compas_timber.connections import LMiterJoint
    from compas_timber.connections import LTenonMortiseJoint
    from compas_timber.connections import TBirdsmouthJoint
    from compas_timber.connections import TButtJoint
    from compas_timber.connections import TDovetailJoint
    from compas_timber.connections import TLapJoint
    from compas_timber.connections import TOliGinaJoint
    from compas_timber.connections import TStepJoint
    from compas_timber.connections import TTenonMortiseJoint
    from compas_timber.connections import XLapJoint
    from compas_timber.connections import XNotchJoint
    from compas_timber.connections import YButtJoint

    return [
        TButtJoint,
        LButtJoint,
        TButtJoint,
        TStepJoint,
        TBirdsmouthJoint,
        LMiterJoint,
        XLapJoint,
        XNotchJoint,
        TLapJoint,
        LLapJoint,
        LFrenchRidgeLapJoint,
        TDovetailJoint,
        BallNodeJoint,
        LTenonMortiseJoint,
        TTenonMortiseJoint,
        YButtJoint,
        TOliGinaJoint,
    ]


def add_gh_param(
    name, io, ghenv, index=None
):  # we could also make beam_names a dict with more info e.g. NickName, Description, Access, hints, etc. this would be defined in joint_options components
    """Adds a parameter to the Grasshopper component.

    Parameters
    ----------
    name : str
        The name of the parameter.
    io : str
        The direction of the parameter. Either "Input" or "Output".
    ghenv : object
        The Grasshopper environment object.

    Returns
    -------
    None

    """
    assert io in ("Output", "Input")
    params = [param.NickName for param in getattr(ghenv.Component.Params, io)]
    if name not in params:
        param = Grasshopper.Kernel.Parameters.Param_GenericObject()
        param.NickName = name
        param.Name = name
        param.Description = name
        param.Access = Grasshopper.Kernel.GH_ParamAccess.item
        param.Optional = True
        if index is None:
            index = getattr(ghenv.Component.Params, io).Count

        registers = dict(Input="RegisterInputParam", Output="RegisterOutputParam")
        getattr(ghenv.Component.Params, registers[io])(param, index)
        ghenv.Component.Params.OnParametersChanged()


def clear_gh_params(ghenv, permanent_param_count=1):
    """Clears all input parameters from the component.

    Parameters
    ----------
    ghenv : object
        The Grasshopper environment object.
    permanent_param_count : int, optional
        The number of parameters that should not be deleted. Default is 1.

    Returns
    -------
    None

    """
    changed = False
    while len(ghenv.Component.Params.Input) > permanent_param_count:
        ghenv.Component.Params.UnregisterInputParameter(ghenv.Component.Params.Input[len(ghenv.Component.Params.Input) - 1], True)
        changed = True
    ghenv.Component.Params.OnParametersChanged()
    return changed


def rename_gh_input(input_name, index, ghenv):
    """Renames a parameter in the Grasshopper component.

    Parameters
    ----------
    ghenv : object
        The Grasshopper environment object.
    input_name : str
        The new name of the parameter.
    index : int
        The index of the parameter to rename.

    Returns
    -------
    None

    """
    param = ghenv.Component.Params.Input[index]
    param.NickName = input_name
    param.Name = input_name
    param.Description = input_name
    ghenv.Component.Params.OnParametersChanged()


def rename_gh_output(output_name, index, ghenv):
    """Renames a parameter in the Grasshopper component.

    Parameters
    ----------
    output_name : str
        The new name of the parameter.
    index : int
        The index of the parameter to rename.
    ghenv : object
        The Grasshopper environment object.

    Returns
    -------
    None

    """
    param = ghenv.Component.Params.Output[index]
    param.NickName = output_name
    param.Name = output_name
    param.Description = output_name
    ghenv.Component.Params.OnParametersChanged()


def manage_dynamic_params(input_names, ghenv, rename_count=0, permanent_param_count=0, keep_connections=True):
    """Clears all input parameters from the component.

    Parameters
    ----------
    input_names : list(str)
        The names of the input parameters.
    ghenv : object
        The Grasshopper environment object.
    rename_count : int, optional
        The number of parameters that should be renamed. Default is 0.
    permanent_param_count : int, optional
        The number of parameters that should not be deleted. Default is 0.

    Returns
    -------
    None

    """
    if not input_names:  # if no names are input
        clear_gh_params(ghenv, permanent_param_count)
        return
    else:
        if keep_connections:
            to_remove = []
            for param in ghenv.Component.Params.Input[permanent_param_count + rename_count :]:
                if param.Name not in input_names:
                    to_remove.append(param)
            for param in to_remove:
                param.IsolateObject()
                ghenv.Component.Params.UnregisterInputParameter(param, True)
            for i, name in enumerate(input_names):
                if i < rename_count:
                    rename_gh_input(name, i + permanent_param_count, ghenv)
                elif name not in [param.Name for param in ghenv.Component.Params.Input]:
                    add_gh_param(name, "Input", ghenv, index=i + permanent_param_count)

        else:
            register_params = False
            if len(ghenv.Component.Params.Input) == len(input_names) + permanent_param_count:  # if param count matches beam_names count
                for i, name in enumerate(input_names):
                    if ghenv.Component.Params.Input[i + permanent_param_count].Name != name:  # if param names don't match
                        register_params = True
                        break
            else:
                register_params = True
            if register_params:
                clear_gh_params(ghenv, permanent_param_count + rename_count)  # we could consider renaming params if we don't want to disconnect GH component inputs
                for i, name in enumerate(input_names):
                    if i < permanent_param_count:
                        continue
                    elif i < rename_count:
                        rename_gh_input(name, i, ghenv)
                    else:
                        add_gh_param(name, "Input", ghenv)


def list_input_valid_cpython(ghenv, Param, name):
    if not Param:
        ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Input parameter %s failed to collect data" % name)
    else:
        if all([_ is None for _ in Param]):
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Input parameter %s failed to collect data" % name)
        elif any([_ is None for _ in Param]):
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Remark, "Input parameter %s contains some Null values" % name)
            return True
        else:
            return True
    return False


def item_input_valid_cpython(ghenv, Param, name):
    if not Param and Param != 0:
        ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Input parameter %s failed to collect data" % name)
    else:
        return True
    return False


def add_cpython_gh_param(
    name, io, ghenv, index=None
):  # we could also make beam_names a dict with more info e.g. NickName, Description, Access, hints, etc. this would be defined in joint_options components
    """Adds a parameter to the Grasshopper component.

    Parameters
    ----------
    name : str
        The name of the parameter.
    io : str
        The direction of the parameter. Either "Input" or "Output".
    ghenv : object
        The Grasshopper environment object.

    Returns
    -------
    None

    """
    assert io in ("Output", "Input")
    existing_param_names = []
    for i in range(len(ghenv.Component.Params.Input)):
        existing_param_names.append(ghenv.Component.Params.Input[i].Name)
    if name not in existing_param_names:
        param = RhinoCodePluginGH.Parameters.ScriptVariableParam()
        param.NickName = name
        param.Name = name
        param.Description = name
        param.Access = Grasshopper.Kernel.GH_ParamAccess.item
        param.Optional = True
        if io == "Input":
            if index is None:
                index = ghenv.Component.Params.Input.Count
            ghenv.Component.Params.RegisterInputParam(param, index)
        else:
            if index is None:
                index = ghenv.Component.Params.Output.Count
            ghenv.Component.Params.RegisterOutputParam(param, index)
        ghenv.Component.VariableParameterMaintenance()
        ghenv.Component.Params.OnParametersChanged()


def clear_cpython_gh_params(ghenv, permanent_param_count=1):
    """Clears all input parameters from the component.

    Parameters
    ----------
    ghenv : object
        The Grasshopper environment object.
    permanent_param_count : int, optional
        The number of parameters that should not be deleted. Default is 1.

    Returns
    -------
    None

    """
    changed = False
    while len(ghenv.Component.Params.Input) > permanent_param_count:
        ghenv.Component.Params.UnregisterInputParameter(ghenv.Component.Params.Input[len(ghenv.Component.Params.Input) - 1], True)
        changed = True
    ghenv.Component.VariableParameterMaintenance()
    ghenv.Component.Params.OnParametersChanged()
    return changed


def rename_cpython_gh_input(input_name, index, ghenv):
    """Renames a parameter in the Grasshopper component.

    Parameters
    ----------
    ghenv : object
        The Grasshopper environment object.
    input_name : str
        The new name of the parameter.
    index : int
        The index of the parameter to rename.

    Returns
    -------
    None

    """
    param = ghenv.Component.Params.Input[index]
    param.NickName = input_name
    param.Name = input_name
    param.Description = input_name
    ghenv.Component.VariableParameterMaintenance()
    ghenv.Component.Params.OnParametersChanged()


def rename_cpython_gh_output(output_name, index, ghenv):
    """Renames a parameter in the Grasshopper component.

    Parameters
    ----------
    output_name : str
        The new name of the parameter.
    index : int
        The index of the parameter to rename.
    ghenv : object
        The Grasshopper environment object.

    Returns
    -------
    None

    """
    param = ghenv.Component.Params.Output[index]
    param.NickName = output_name
    param.Name = output_name
    param.Description = output_name
    ghenv.Component.VariableParameterMaintenance()
    ghenv.Component.Params.OnParametersChanged()


def manage_cpython_dynamic_params(input_names, ghenv, rename_count=0, permanent_param_count=0, keep_connections=True):
    """Clears all input parameters from the component.

    Parameters
    ----------
    input_names : list(str)
        The names of the input parameters.
    ghenv : object
        The Grasshopper environment object.
    rename_count : int, optional
        The number of parameters that should be renamed. Default is 0.
    permanent_param_count : int, optional
        The number of parameters that should not be deleted. Default is 0.

    Returns
    -------
    None

    """
    if not input_names:  # if no names are input
        clear_cpython_gh_params(ghenv, permanent_param_count)
        return
    else:
        if keep_connections:
            to_remove = []
            for i in range((permanent_param_count + rename_count), len(ghenv.Component.Params.Input)):
                if ghenv.Component.Params.Input[i].Name not in input_names:
                    to_remove.append(ghenv.Component.Params.Input[i])
            for param in to_remove:
                ghenv.Component.Params.UnregisterInputParameter(param, True)

            for i, name in enumerate(input_names):
                if i < rename_count:
                    rename_cpython_gh_input(name, i + permanent_param_count, ghenv)
                else:
                    add_cpython_gh_param(name, "Input", ghenv, index=i + permanent_param_count)

        else:
            register_params = False
            if len(ghenv.Component.Params.Input) == len(input_names) + permanent_param_count:  # if param count matches beam_names count
                for i, name in enumerate(input_names):
                    if ghenv.Component.Params.Input[i + permanent_param_count].Name != name:  # if param names don't match
                        register_params = True
                        break
            else:
                register_params = True
            if register_params:
                clear_cpython_gh_params(ghenv, permanent_param_count + rename_count)  # we could consider renaming params if we don't want to disconnect GH component inputs
                for i, name in enumerate(input_names):
                    if i < permanent_param_count:
                        continue
                    elif i < rename_count:
                        rename_cpython_gh_input(name, i, ghenv)
                    else:
                        add_cpython_gh_param(name, "Input", ghenv, index=i + permanent_param_count)
    ghenv.Component.VariableParameterMaintenance()



# r: compas_timber>=1.0.0
# flake8: noqa
import Grasshopper
import System
import warnings

from compas_timber.planning import BeamNester as CTBeamNester
from compas_timber.planning import BeamStock
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class BeamNester(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, model, stock_catalog: System.Collections.Generic.List[object], spacing: float, fast: bool):
        if not item_input_valid_cpython(ghenv, model, "Model"):
            return
        if not item_input_valid_cpython(ghenv, stock_catalog, "Stock Catalog"):
            return

        # Validate stock catalog items
        stock_catalog = list(stock_catalog)
        for stock in stock_catalog:
            if not isinstance(stock, BeamStock):
                ghenv.Component.AddRuntimeMessage(
                    Grasshopper.Kernel.GH_RuntimeMessageLevel.Error,
                    f"All items in Stock Catalog must be of type Stock. Found {type(stock)}",
                )
                return

        fast = fast or False
        spacing = spacing or 0.0
        # Catch any warnings raised for unnested beams
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            beam_nester = CTBeamNester(model, stock_catalog, spacing=spacing)
            nesting_result = beam_nester.nest(fast=fast)
            nesting_summary = nesting_result.summary
            if w:
                for warning in w:
                    ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, str(warning.message))
                # Add warnings section to summary
                nesting_summary += "\n\nWARNINGS:\n~~~~~~"
                for i, warning in enumerate(w, 1):
                    nesting_summary += f"\nWarning_{i}:\n{str(warning.message)}\n--------"

        return nesting_result, nesting_summary



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper

from compas.scene import SceneObject
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class ShowFeatureErrors(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, debug_info, i: int):
        if not item_input_valid_cpython(ghenv, i, "i"):
            i = 0
        if not item_input_valid_cpython(ghenv, debug_info, "DebugInfo"):
            return
        if not debug_info.feature_errors:
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "No feature errors found in input parameter 'DebugInfo'")
            return
        feature_errors = debug_info.feature_errors
        i = int(i) % len(feature_errors)
        error = feature_errors[i]

        geometries = [error.feature_geometry, error.element_geometry]
        geometries = flatten(geometries)
        geo_objs = [SceneObject(item=geo) for geo in geometries]
        output = []
        for obj in geo_objs:
            output.extend(obj.draw())

        return error.message, output


def flatten(lst):
    if not lst:
        return lst
    if isinstance(lst[0], list):
        return flatten(lst[0]) + flatten(lst[1:])
    return lst[:1] + flatten(lst[1:])



# r: timber_design>=0.1.0
# venv: td_migration
"""Generates a direct joint between two elements. This overrides other joint rules."""

import inspect

import Grasshopper  # type: ignore
from compas_timber.connections import PlateJoint
from System.Windows.Forms import ToolStripSeparator

from timber_design.ghpython import get_leaf_subclasses
from timber_design.ghpython import item_input_valid_cpython
from timber_design.ghpython import manage_cpython_dynamic_params
from timber_design.ghpython import rename_cpython_gh_output
from timber_design.ghpython import warning
from timber_design.workflow import DirectRule


class DirectPlateJointRule(Grasshopper.Kernel.GH_ScriptInstance):
    def __init__(self):
        super(DirectPlateJointRule, self).__init__()
        self.classes = {}
        for cls in get_leaf_subclasses(PlateJoint):
            if cls.MAX_ELEMENT_COUNT == 2:
                self.classes[cls.__name__] = cls

        self.joint_type = self.classes.get(self.component.Params.Output[0].NickName, None)

    @property
    def component(self):
        return ghenv.Component  # type: ignore  # noqa: F821

    def RunScript(self, *args):
        if not self.joint_type:
            self.component.Message = "Select joint type from context menu (right click)"
            warning(self.component, "Select joint type from context menu (right click)")
            return None

        self.component.Message = self.joint_type.__name__
        plate_a = args[0]
        plate_b = args[1]

        if not item_input_valid_cpython(ghenv, plate_a, self.arg_names()[0]) or not item_input_valid_cpython(ghenv, plate_b, self.arg_names()[1]):
            return

        kwargs = {}
        for i, val in enumerate(args[2:]):
            if val is not None:
                kwargs[self.arg_names()[i + 2]] = val

        return DirectRule(self.joint_type, [plate_a, plate_b], **kwargs)

    def arg_names(self):
        return inspect.getargspec(self.joint_type.__init__)[0][1:3] + ["max_distance"]

    def AppendAdditionalMenuItems(self, menu):
        for name in self.classes.keys():
            item = menu.Items.Add(name, None, self.on_item_click)
            if self.joint_type and name == self.joint_type.__name__:
                item.Checked = True
        menu.Items.Add(ToolStripSeparator())

    def on_item_click(self, sender, event_info):
        self.joint_type = self.classes[str(sender)]
        rename_cpython_gh_output(self.joint_type.__name__, 0, ghenv)
        manage_cpython_dynamic_params(self.arg_names(), ghenv, rename_count=2, permanent_param_count=0)
        self.component.ExpireSolution(True)



import math

from compas.geometry import Frame
from compas.geometry import Line
from compas.geometry import Plane
from compas.geometry import Point
from compas.geometry import Polyline
from compas.geometry import Vector
from compas.geometry import angle_vectors
from compas.geometry import angle_vectors_signed
from compas.geometry import bounding_box_xy
from compas.geometry import closest_point_on_plane
from compas.geometry import closest_point_on_segment
from compas.geometry import cross_vectors
from compas.geometry import distance_point_point_sqrd
from compas.geometry import dot_vectors
from compas.geometry import intersection_line_line
from compas.geometry import intersection_line_segment
from compas.geometry import matrix_from_frame_to_frame
from compas.geometry import offset_line
from compas.geometry import offset_polyline
from compas.tolerance import Tolerance
from compas_timber.connections import ConnectionSolver
from compas_timber.connections import JointTopology
from compas_timber.connections import LButtJoint
from compas_timber.connections import TButtJoint
from compas_timber.elements import Beam
from compas_timber.elements import Plate
from compas_timber.fabrication import FreeContour
from compas_timber.model import TimberModel

from timber_design.workflow import CategoryRule


class SurfaceModel(object):
    """Create a timber assembly from a surface.

    Parameters
    ----------
    surface : :class:`compas.geometry.Surface`
        The surface to create the assembly from. must be planar.
    beam_width : float
        The height of the beams aka thickness of wall cavity normal to the surface.
    frame_depth : float
        The width of the beams.
    stud_spacing : float
        The spacing between the studs.
    z_axis : :class:`compas.geometry.Vector`, optional
        Determines the orientation of the posts inside the frame.
        Default is ``Vector.Zaxis``.


    Attributes
    ----------
    beams : list of :class:`compas_timber.elements.Beam`
        The beams of the assembly.
    rules : list of :class:`compas_timber.design.CategoryRule`
        The rules for the assembly.
    centerlines : list of :class:`compas.geometry.Line`
        The centerlines of the beams.
    normal : :class:`compas.geometry.Vector`
        The normal of the surface.
    panel_length : float
        The length of the panel.
    panel_height : float
        The height of the panel.
    frame : :class:`compas.geometry.Frame`
        The frame of the assembly.
    jack_studs : list of :class:`compas_timber.model.SurfaceAssembly.BeamDefinition`
        The jack studs of the assembly.
    king_studs : list of :class:`compas_timber.model.SurfaceAssembly.BeamDefinition`
        The king studs of the assembly.
    edge_studs : list of :class:`compas_timber.model.SurfaceAssembly.BeamDefinition`
        The edge studs of the assembly.
    studs : list of :class:`compas_timber.model.SurfaceAssembly.BeamDefinition`
        The studs of the assembly.
    sills : list of :class:`compas_timber.model.SurfaceAssembly.BeamDefinition`
        The sills of the assembly.
    headers : list of :class:`compas_timber.model.SurfaceAssembly.BeamDefinition`
        The headers of the assembly.
    plates : list of :class:`compas_timber.model.SurfaceAssembly.BeamDefinition`

    """

    BEAM_CATEGORY_NAMES = ["stud", "king_stud", "jack_stud", "edge_stud", "plate", "header", "sill"]

    def __init__(
        self,
        surface,
        stud_spacing,
        beam_width=None,
        frame_depth=None,
        z_axis=None,
        tolerance=Tolerance(unit="MM", absolute=1e-3, relative=1e-3),
        sheeting_outside=None,
        sheeting_inside=None,
        lintel_posts=True,
        edge_stud_offset=0.0,
        custom_dimensions=None,
        joint_overrides=None,
    ):
        self.surface = surface
        self.beam_width = beam_width
        self.frame_depth = frame_depth
        self.stud_spacing = stud_spacing
        self._z_axis = z_axis or Vector.Zaxis()
        self.sheeting_outside = sheeting_outside
        self.sheeting_inside = sheeting_inside
        self.edge_stud_offset = edge_stud_offset or 0.0
        self.lintel_posts = lintel_posts
        self._normal = None
        self.outer_polyline = None
        self.inner_polylines = []
        self.edges = []
        self._frame = None
        self._panel_length = None
        self._panel_height = None
        self._elements = []
        self._beam_definitions = []
        self._rules = []
        self.windows = []
        self._features = []
        self.beam_dimensions = {}
        self.joint_overrides = joint_overrides
        self.dist_tolerance = tolerance.relative

        for key in self.BEAM_CATEGORY_NAMES:
            self.beam_dimensions[key] = [self.beam_width, self.frame_depth]
        if custom_dimensions:
            for key, value in custom_dimensions.items():
                if value:
                    if value[0] != 0:
                        self.beam_dimensions[key][0] = value[0]
                    if value[1] != 0:
                        self.beam_dimensions[key][1] = value[1]
        self.parse_loops()
        self.generate_frame()
        self.generate_plates()

    @property
    def z_axis(self):
        cross = cross_vectors(self.normal, self._z_axis)
        return Vector(*cross_vectors(cross, self.normal))

    @property
    def default_rules(self):
        return [
            (CategoryRule(LButtJoint, "edge_stud", "plate") if self.edge_stud_offset == 0 else CategoryRule(TButtJoint, "edge_stud", "plate")),
            CategoryRule(TButtJoint, "stud", "plate"),
            CategoryRule(TButtJoint, "stud", "header"),
            CategoryRule(TButtJoint, "stud", "sill"),
            CategoryRule(LButtJoint, "jack_stud", "plate"),
            CategoryRule(TButtJoint, "jack_stud", "plate"),
            CategoryRule(LButtJoint, "jack_stud", "header"),
            CategoryRule(TButtJoint, "jack_stud", "header"),
            CategoryRule(TButtJoint, "king_stud", "plate"),
            CategoryRule(LButtJoint, "king_stud", "plate"),
            CategoryRule(TButtJoint, "king_stud", "sill"),
            CategoryRule(TButtJoint, "king_stud", "header"),
            CategoryRule(TButtJoint, "sill", "jack_stud"),
        ]

    @property
    def rules(self):
        if not self._rules:
            self._rules = self.default_rules
            if self.joint_overrides:
                for rule in self.joint_overrides:
                    rule_set = set([rule.category_a, rule.category_b])
                    for i, _rule in enumerate(self._rules):
                        _set = set([_rule.category_a, _rule.category_b])
                        if rule_set == _set:
                            self._rules[i] = rule
                            break
        return self._rules

    @property
    def centerlines(self):
        return [beam_def.centerline for beam_def in self.beam_definitions]

    @property
    def elements(self):
        elements = []
        for window in self.windows:
            self._beam_definitions.extend(window._beam_definitions)
        for beam_def in self._beam_definitions:
            elements.append(beam_def.to_beam())
        elements.extend(self._elements)
        return elements

    @property
    def features(self):
        return self._features

    @property
    def plate_elements(self):
        for plate in self._elements:
            if isinstance(plate, Plate):
                yield plate

    @property
    def points(self):
        return self.outer_polyline.points

    @property
    def normal(self):
        if not self._normal:
            self._normal = self.surface.native_brep.Faces[0].NormalAt(0.5, 0.5)
        return self._normal

    @property
    def panel_length(self):
        if not self._panel_length:
            _ = self.frame
        return self._panel_length

    @property
    def panel_height(self):
        if not self._panel_height:
            _ = self.frame
        return self._panel_height

    @property
    def frame(self):
        if not self._frame:
            self._frame, self._panel_length, self._panel_height = get_frame(self.points, self.normal, self.z_axis)
        return self._frame

    @property
    def jack_studs(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "jack_stud"]

    @property
    def king_studs(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "king_stud"]

    @property
    def edge_studs(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "edge_stud"]

    @property
    def studs(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "stud"]

    @property
    def sills(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "sill"]

    @property
    def headers(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "header"]

    @property
    def plates(self):
        return [beam_def for beam_def in self._beam_definitions if beam_def.type == "plate"]

    @classmethod
    def beam_category_names(cls):
        return SurfaceModel.BEAM_CATEGORY_NAMES

    def generate_frame(self):
        self.generate_perimeter_beams()
        self.generate_windows()
        self.generate_studs()

    def create_model(self):
        model = TimberModel()
        for element in self.elements:
            model.add_element(element)
        topologies = []
        solver = ConnectionSolver()
        found_pairs = solver.find_intersecting_pairs(list(model.beams), rtree=True, max_distance=self.dist_tolerance)
        for pair in found_pairs:
            beam_a, beam_b = pair
            detected_topo, beam_a, beam_b = solver.find_topology(beam_a, beam_b, max_distance=self.dist_tolerance)
            if not detected_topo == JointTopology.TOPO_UNKNOWN:
                topologies.append({"detected_topo": detected_topo, "beam_a": beam_a, "beam_b": beam_b})
                for rule in self.rules:
                    if not rule.comply(pair):
                        continue
                    if rule.joint_type.SUPPORTED_TOPOLOGY != detected_topo:
                        continue
                    else:
                        if rule.joint_type == LButtJoint:
                            beam_a, beam_b = rule.reorder([beam_a, beam_b])
                        rule.joint_type.create(model, beam_a, beam_b, **rule.kwargs)
        model.set_topologies(topologies)
        return model

    def parse_loops(self):
        for loop in self.surface.loops:
            polyline_points = []
            polyline_length = 0.0
            for i, edge in enumerate(loop.edges):
                if not edge.is_line:
                    raise ValueError("function only supprorts polyline edges")
                if edge.start_vertex.point in [
                    loop.edges[i - 1].start_vertex.point,
                    loop.edges[i - 1].end_vertex.point,
                ]:
                    polyline_points.append(edge.start_vertex.point)
                else:
                    polyline_points.append(edge.end_vertex.point)
                polyline_length += edge.length
            polyline_points.append(polyline_points[0])
            loop_polyline = Polyline(polyline_points)
            offset_dist = self.dist_tolerance * 1000
            if loop.is_outer:
                offset_loop = Polyline(offset_polyline(loop_polyline, offset_dist, self.normal))
                if offset_loop.length > loop_polyline.length:
                    polyline_points.reverse()
                self.outer_polyline = Polyline(polyline_points)
            else:
                offset_loop = Polyline(offset_polyline(loop_polyline, offset_dist, self.normal))
                if offset_loop.length < loop_polyline.length:
                    polyline_points.reverse()
                self.inner_polylines.append(Polyline(polyline_points))

    def generate_perimeter_beams(self):
        interior_indices = self.get_interior_segment_indices(self.outer_polyline)
        for i, segment in enumerate(self.outer_polyline.lines):
            beam_def = self.BeamDefinition(segment, parent=self)
            if i in interior_indices:
                if angle_vectors(segment.direction, self.z_axis, deg=True) < 45 or angle_vectors(segment.direction, self.z_axis, deg=True) > 135:
                    if self.lintel_posts:
                        beam_def.type = "jack_stud"
                    else:
                        beam_def.type = "king_stud"
                else:
                    beam_def.type = "header"
            else:
                if angle_vectors(segment.direction, self.z_axis, deg=True) < 45 or angle_vectors(segment.direction, self.z_axis, deg=True) > 135:
                    beam_def.type = "edge_stud"
                else:
                    beam_def.type = "plate"
            self._beam_definitions.append(beam_def)
        self._beam_definitions = self.offset_elements(self._beam_definitions)
        if self.lintel_posts:
            for beam_def in self._beam_definitions:
                if beam_def.type == "jack_stud":
                    offset = (self.beam_dimensions["jack_stud"][0] + self.beam_dimensions["king_stud"][0]) / 2
                    king_line = offset_line(beam_def.centerline, offset, self.normal)
                    self._beam_definitions.append(self.BeamDefinition(king_line, type="king_stud", parent=self))

    def get_interior_segment_indices(self, polyline):
        points = polyline.points[0:-1]
        out = []
        for index in range(len(points)):
            if index == 0:
                angle = angle_vectors_signed(points[-1] - points[0], points[1] - points[0], self.normal, deg=True)
            elif index == len(points) - 1:
                angle = angle_vectors_signed(points[-2] - points[-1], points[0] - points[-1], self.normal, deg=True)
            else:
                angle = angle_vectors_signed(points[index - 1] - points[index], points[index + 1] - points[index], self.normal, deg=True)
            if angle > 0:
                out.append(index)
        if len(out) > 0:
            out.insert(0, out[0] - 1)
        return set(out)

    def offset_elements(self, element_loop):
        offset_loop = []
        for beam_def in element_loop:
            beam_def.offset(self.beam_dimensions[beam_def.type][0] / 2)
            if beam_def.type == "edge_stud":
                beam_def.offset(self.edge_stud_offset)
            offset_loop.append(beam_def)

        for i, beam_def in enumerate(offset_loop):
            if self.edge_stud_offset > 0:
                if beam_def.type != "plate":
                    element_before = offset_loop[i - 1]
                    element_after = offset_loop[(i + 1) % len(offset_loop)]
                    start_point = intersection_line_line(beam_def.centerline, element_before.centerline, self.dist_tolerance)[0]
                    end_point = intersection_line_line(beam_def.centerline, element_after.centerline, self.dist_tolerance)[0]
                    if start_point and end_point:
                        beam_def.centerline = Line(start_point, end_point)
                    else:
                        raise ValueError("edges are parallel, no intersection found")
            else:
                element_before = offset_loop[i - 1]
                element_after = offset_loop[(i + 1) % len(offset_loop)]
                start_point, _ = intersection_line_line(beam_def.centerline, element_before.centerline, self.dist_tolerance)
                end_point, _ = intersection_line_line(beam_def.centerline, element_after.centerline, self.dist_tolerance)
                if start_point and end_point:
                    beam_def.centerline = Line(start_point, end_point)
        return offset_loop

    def generate_windows(self):
        for polyline in self.inner_polylines:
            self.windows.append(self.Window(polyline, parent=self))
            self._beam_definitions.extend(self.windows[-1]._beam_definitions)

    def generate_studs(self):
        self.generate_stud_lines()
        self.trim_jack_studs()
        self.trim_king_studs()
        self.trim_studs()
        self.cull_overlaps()

    def generate_stud_lines(self):
        x_position = self.stud_spacing
        while x_position < self.panel_length - self.beam_width:
            start_point = Point(x_position, 0, 0)
            start_point.transform(matrix_from_frame_to_frame(Frame.worldXY(), self.frame))
            line = Line.from_point_and_vector(start_point, self.z_axis * self.panel_height)
            self._beam_definitions.append(self.BeamDefinition(line, type="stud", parent=self))
            x_position += self.stud_spacing

    def get_beam_intersections(self, beam_def, *element_lists_to_intersect):
        intersections = []
        dots = []
        for element_list in element_lists_to_intersect:
            for element_to_intersect in element_list:
                point = intersection_line_segment(beam_def.z_aligned_centerline, element_to_intersect.centerline, 0.01)[0]
                if point:
                    intersections.append(point)
        if len(intersections) > 1:
            intersections.sort(key=lambda x: dot_vectors(x, self.z_axis))
            dots = [dot_vectors(Vector.from_start_end(beam_def.z_aligned_centerline.start, x), self.z_axis) / beam_def.centerline.length for x in intersections]
        return intersections, dots

    def trim_jack_studs(self):
        for beam_def in self.jack_studs:
            intersections, dots = self.get_beam_intersections(beam_def, self.plates, self.headers)
            if len(intersections) > 1:
                bottom = None
                for i, dot in enumerate(dots):
                    if dot < 0:
                        bottom = intersections[i]
                top = beam_def.z_aligned_centerline.end
                if not bottom:
                    bottom = beam_def.z_aligned_centerline.start
            beam_def.set_centerline(Line(bottom, top))

    def trim_king_studs(self):
        for beam_def in self.king_studs:
            intersections, dots = self.get_beam_intersections(beam_def, self.plates, self.headers, self.sills)
            if len(intersections) > 1:
                bottom, top = None, None
                for i, dot in enumerate(dots):
                    if dot < -0.01:
                        bottom = intersections[i]  # last intersection below sill
                    if dot > 1.01:
                        top = intersections[i]  # first intersection above header
                        break
                if not bottom:
                    bottom = beam_def.z_aligned_centerline.start
                if not top:
                    top = beam_def.z_aligned_centerline.end
                beam_def.set_centerline(Line(bottom, top))

    def trim_studs(self):
        stud_elements = []
        while len(self.studs) > 0:
            for beam_def in self._beam_definitions:
                if beam_def.type == "stud":
                    intersections, _ = self.get_beam_intersections(beam_def, self.plates, self.headers, self.sills)
                    while len(intersections) > 1:
                        top = intersections.pop()
                        bottom = intersections.pop()
                        stud_elements.append(self.BeamDefinition(Line(bottom, top), type="stud", parent=self))
                    self._beam_definitions.remove(beam_def)
        self._beam_definitions.extend(stud_elements)

    def cull_overlaps(self):
        for beam_def in self.studs:
            for other_element in self.king_studs + self.jack_studs + self.edge_studs:
                if self.distance_between_elements(beam_def, other_element) < (self.beam_dimensions[beam_def.type][0] + self.beam_dimensions[other_element.type][0]) / 2:
                    self._beam_definitions.remove(beam_def)
                    break

    def distance_between_elements(self, element_one, element_two):
        distances = []
        for pt in element_one.centerline:
            cp = closest_point_on_segment(pt, element_two.centerline)
            distances.append(distance_point_point_sqrd(pt, cp))
        for pt in element_two.centerline:
            cp = closest_point_on_segment(pt, element_one.centerline)
            distances.append(distance_point_point_sqrd(pt, cp))
        return math.sqrt(min(distances))

    def generate_plates(self):
        if self.sheeting_inside:
            self._elements.append(Plate(self.outer_polyline, self.sheeting_inside))
        if self.sheeting_outside:
            pline = self.outer_polyline.copy()
            pline.translate(self.frame.zaxis * (self.frame_depth + self.sheeting_outside))
            self._elements.append(Plate(pline, self.sheeting_outside))
        for window in self.windows:
            for plate in self.plate_elements:
                window.apply_contour_to_plate(plate)

    class Window(object):
        """
        A window object for the SurfaceAssembly.

        Parameters
        ----------
        outline : :class:`compas.geometry.Polyline`
            The outline of the window.
        sill_height : float, optional
            The height of the sill.
        header_height : float, optional
            The height of the header.
        parent : :class:`compas_timber.model.SurfaceAssembly`
            The parent of the window.

        Attributes
        ----------
        outline : :class:`compas.geometry.Polyline`
            The outline of the window.
        sill_height : float
            The height of the sill.
        header_height : float
            The height of the header.
        parent : :class:`compas_timber.model.SurfaceAssembly`
            The parent of the window.
        z_axis : :class:`compas.geometry.Vector`
            The z axis of the parent.
        normal : :class:`compas.geometry.Vector`
            The normal of the parent.
        beam_dimensions : dict
            The beam dimensions of the parent.
        beam_definions : list of :class:`compas_timber.model.SurfaceAssembly.BeamDefinition`
            The beam_definions of the window.
        length : float
            The length of the window.
        height : float
            The height of the window.
        frame : :class:`compas.geometry.Frame`
            The frame of the window.

        """

        def __init__(self, outline, sill_height=None, header_height=None, parent=None):
            self.outline = outline
            if sill_height:
                self.sill_height = sill_height
            else:
                self.sill_height = parent.beam_dimensions["sill"][0]
            if header_height:
                self.header_height = header_height
            else:
                self.header_height = parent.beam_dimensions["header"][0]
            self.parent = parent
            self.z_axis = parent.frame.yaxis
            self.normal = parent.frame.zaxis
            self.beam_dimensions = parent.beam_dimensions
            self._beam_definitions = []
            self._length = None
            self._height = None
            self._frame = None
            self.dist_tolerance = parent.dist_tolerance
            self.process_outlines()

        @property
        def jack_studs(self):
            return [beam_def for beam_def in self._beam_definitions if beam_def.type == "jack_stud"]

        @property
        def sills(self):
            return [beam_def for beam_def in self._beam_definitions if beam_def.type == "sill"]

        @property
        def headers(self):
            return [beam_def for beam_def in self._beam_definitions if beam_def.type == "header"]

        @property
        def length(self):
            if not self._length:
                _ = self.frame
            return self._length

        @property
        def height(self):
            if not self._height:
                _ = self.frame
            return self._height

        @property
        def frame(self):
            if not self._frame:
                self._frame, self._panel_length, self._panel_height = get_frame(self.points, self.parent.normal, self.zaxis)
            return self._frame

        def apply_contour_to_plate(self, plate):
            projected_points = []
            for point in self.outline.points:
                projected_points.append(closest_point_on_plane(point, Plane.from_frame(plate.frame)))
            feature = FreeContour.from_polyline_and_element(Polyline(projected_points), plate)
            plate.add_feature(feature)

        def process_outlines(self):
            for i, segment in enumerate(self.outline.lines):
                beam_def = SurfaceModel.BeamDefinition(segment, parent=self)
                if angle_vectors(segment.direction, self.z_axis, deg=True) < 1 or angle_vectors(segment.direction, self.z_axis, deg=True) > 179:
                    if self.parent.lintel_posts:
                        beam_def.type = "jack_stud"
                    else:
                        beam_def.type = "king_stud"
                else:
                    ray = Line.from_point_and_vector(segment.point_at(0.5), self.z_axis)
                    pts = []
                    for seg in self.outline.lines:
                        if seg != segment:
                            pt = intersection_line_segment(ray, seg, self.dist_tolerance)[0]
                            if pt:
                                pts.append(Point(*pt))
                    if len(pts) > 1:
                        raise ValueError("Window outline is wonky")
                    elif len(pts) == 1:
                        vector = Vector.from_start_end(ray.start, pts[0])
                        if dot_vectors(vector, self.z_axis) < 0:
                            beam_def.type = "header"
                        else:
                            beam_def.type = "sill"
                self._beam_definitions.append(beam_def)
            self._beam_definitions = self.parent.offset_elements(self._beam_definitions)
            if self.parent.lintel_posts:
                for beam_def in self.jack_studs:
                    offset = (self.parent.beam_dimensions["jack_stud"][0] + self.parent.beam_dimensions["king_stud"][0]) / 2
                    king_line = offset_line(beam_def.centerline, offset, self.normal)
                    self._beam_definitions.append(self.parent.BeamDefinition(king_line, type="king_stud", parent=self))

    class BeamDefinition(object):
        """
        Container for Beam attributes before beam is instantiated.

        Parameters
        ----------
        centerline : :class:`compas.geometry.Line`
            The centerline of the beam.
        width : float, optional
            The width of the beam.
        height : float, optional
            The height of the beam.
        z_axis : :class:`compas.geometry.Vector`, optional
            The z axis of the beam.
        normal : :class:`compas.geometry.Vector`, optional
            The normal of the beam.
        type : str, optional
            The type of the beam.
        polyline : :class:`compas.geometry.Polyline`, optional
            The polyline of the beam.
        parent : :class:`compas_timber.model.SurfaceAssembly` or :class:`compas_timber.model.SurfaceAssembly.Window`
            The parent of the beam.

        Attributes
        ----------
        centerline : :class:`compas.geometry.Line`
            The centerline of the beam element.
        width : float
            The width of the beam element.
        height : float
            The height of the beam element.
        z_axis : :class:`compas.geometry.Vector`
            The z axis of the parent (Not the Beam).
        normal : :class:`compas.geometry.Vector`
            The normal of the parent.


        """

        def __init__(
            self,
            centerline,
            width=None,
            height=None,
            z_axis=None,
            normal=None,
            type=None,
            parent=None,
        ):
            self.original_centerline = centerline
            self.centerline = Line(centerline[0], centerline[1])
            self._width = width
            self._height = height
            self._normal = normal
            self.type = type
            self.parent = parent

        @property
        def width(self):
            return self._width if self._width else self.parent.beam_dimensions[self.type][0]

        @property
        def height(self):
            return self._height if self._height else self.parent.beam_dimensions[self.type][1]

        @property
        def z_aligned_centerline(self):
            if dot_vectors(self.centerline.direction, self.parent.z_axis) < 0:
                return Line(self.centerline.end, self.centerline.start)
            else:
                return self.centerline

        @property
        def normal(self):
            return self._normal if self._normal else self.parent.normal

        def offset(self, distance):
            line = offset_line(self.centerline, distance, self.normal)
            self.centerline = Line(line[0], line[1])

        def translate(self, vector):
            self.centerline.transform(vector)

        def to_beam(self):
            centerline = Line(*self.centerline)
            centerline.translate(self.normal * 0.5 * self.height)
            beam = Beam.from_centerline(centerline, self.width, self.height, self.normal)
            beam.attributes["category"] = self.type
            return beam

        def set_centerline(self, line):
            self.centerline = line


def get_frame(points, normal, z_axis):
    frame = Frame(points[0], cross_vectors(z_axis, normal), z_axis)
    pts_on_xy = [Point(point.x, point.y, point.z) for point in points]
    for point in pts_on_xy:
        point.transform(matrix_from_frame_to_frame(frame, Frame.worldXY()))
    box_min, _, box_max, _ = bounding_box_xy(pts_on_xy)
    pt = Point(box_min[0], box_min[1], box_min[2])
    box_length = box_max[0] - box_min[0]
    box_height = box_max[1] - box_min[1]
    pt.transform(matrix_from_frame_to_frame(Frame.worldXY(), frame))
    frame.point = pt
    return frame, box_length, box_height
